"""The main TAMS class."""

import argparse
import datetime
import logging
from pathlib import Path
from typing import Any
import numpy as np
import numpy.typing as npt
import toml
from pytams.database import Database
from pytams.taskrunner import get_runner_type
from pytams.utils import get_min_scored
from pytams.utils import setup_logger
from pytams.worker import ms_worker
from pytams.worker import pool_worker

_logger = logging.getLogger(__name__)

STALL_TOL = 1e-10


def parse_cl_args(a_args: list[str] | None = None) -> argparse.Namespace:
    """Parse provided list or default CL argv.

    Args:
        a_args: optional list of options
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", help="pyTAMS input .toml file", default="input.toml")
    return parser.parse_args() if a_args is None else parser.parse_args(a_args)


class TAMS:
    """A class implementing TAMS.

    The interface to TAMS, implementing the main steps of
    the algorithm.

    Initialization of the TAMS class requires a forward model
    type which encapsulate all the model-specific code, and
    an optional list of options.

    The algorithm is roughly divided in two steps:
    1. Initialization of the trajectory ensemble
    2. Splitting iterations

    Separate control of the parallelism is provided for
    both steps.

    All the algorithm data are contained in the TAMS database.
    For control purposes, a walltime limit is also provided. It is
    passed to working and lead to the termination of the algorithm
    in a state that can be saved to disk and restarted at a later stage.

    Attributes:
        _fmodel_t: the forward model type
        _parameters: the dictionary of parameters
        _wallTime: the walltime limit
        _startDate: the date the algorithm started
        _plot_diags: whether or not to plot diagnostics during splitting iterations
        _init_ensemble_only: whether or not to stop after initializing the trajectory ensemble
        _tdb: the trajectory database (containing all trajectories)
    """

    def __init__(self, fmodel_t: Any, a_args: list[str] | None = None) -> None:
        """Initialize a TAMS object.

        Args:
            fmodel_t: the forward model type
            a_args: optional list of options

        Raises:
            ValueError: if the input file is not found
        """
        self._fmodel_t = fmodel_t

        input_file = vars(parse_cl_args(a_args=a_args))["input"]
        if not Path(input_file).exists():
            err_msg = f"Could not find the {input_file} TAMS input file !"
            _logger.exception(err_msg)
            raise ValueError(err_msg)

        with Path(input_file).open("r") as f:
            self._parameters = toml.load(f)

        # Setup logger
        setup_logger(self._parameters)

        # Parse user-inputs
        tams_subdict = self._parameters["tams"]
        if "ntrajectories" not in tams_subdict or "nsplititer" not in tams_subdict:
            err_msg = "TAMS 'ntrajectories' and 'nsplititer' must be specified in the input file !"
            _logger.exception(err_msg)
            raise ValueError

        n_traj: int = tams_subdict.get("ntrajectories")
        n_split_iter: int = tams_subdict.get("nsplititer")
        self._wallTime: float = tams_subdict.get("walltime", 24.0 * 3600.0)
        self._plot_diags = tams_subdict.get("diagnostics", False)
        self._init_ensemble_only = tams_subdict.get("init_ensemble_only", False)

        # Database
        self._tdb = Database(fmodel_t, self._parameters, n_traj, n_split_iter, read_only=False)
        self._tdb.load_data()

        # Time management uses UTC date
        # to make sure workers are always in sync
        self._startDate: datetime.datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        self._endDate: datetime.datetime = self._startDate + datetime.timedelta(seconds=self._wallTime)

        # Initialize an empty trajectory ensemble
        if self._tdb.is_empty():
            self._tdb.init_active_ensemble()

    def n_traj(self) -> int:
        """Return the number of trajectory used for TAMS.

        Note that this is the requested number of trajectory, not
        the current length of the trajectory ensemble.

        Return:
            number of trajectory
        """
        return self._tdb.n_traj()

    def elapsed_time(self) -> float:
        """Return the elapsed wallclock time.

        Since the initialization of the TAMS object [seconds].

        Returns:
           TAMS elapse time.

        Raises:
            ValueError: if the start date is not set
        """
        delta: datetime.timedelta = datetime.datetime.now(tz=datetime.timezone.utc) - self._startDate
        if delta:
            return delta.total_seconds()

        err_msg = "TAMS start date is not set !"
        _logger.exception(err_msg)
        raise ValueError

    def remaining_walltime(self) -> float:
        """Return the remaining wallclock time.

        [seconds]

        Returns:
           TAMS remaining wall time.
        """
        return self._wallTime - self.elapsed_time()

    def out_of_time(self) -> bool:
        """Return true if insufficient walltime remains.

        Allows for 5% slack to allows time for workers to finish
        their work (especially with Dask+Slurm backend).

        Returns:
           boolean indicating wall time availability.
        """
        return self.remaining_walltime() < 0.05 * self._wallTime

    def generate_trajectory_ensemble(self) -> None:
        """Schedule the generation of an ensemble of stochastic trajectories.

        Loop over all the trajectories in the database and schedule
        advancing them to either end time or convergence with the
        runner.

        The runner will use the number of workers specified in the
        input file under the runner section.

        Raises:
            Error if the runner fails
        """
        inf_msg = f"Creating the initial ensemble of {self._tdb.n_traj()} trajectories"
        _logger.info(inf_msg)

        with get_runner_type(self._parameters)(
            self._parameters, pool_worker, self._parameters.get("runner", {}).get("nworker_init", 1)
        ) as runner:
            for t in self._tdb.traj_list():
                task = [t, self._endDate, self._tdb.pool_file(), self._tdb.path()]
                runner.make_promise(task)

            try:
                t_list = runner.execute_promises()
            except:
                err_msg = f"Failed to generate the initial ensemble of {self._tdb.n_traj()} trajectories"
                _logger.exception(err_msg)
                raise

        # Re-order list since runner does not guarantee order
        # And update list of trajectories in the database
        t_list.sort(key=lambda t: t.id())
        self._tdb.update_traj_list(t_list)

        if self._tdb.count_ended_traj() == self._tdb.n_traj():
            self._tdb.set_init_ensemble_flag(True)

        inf_msg = f"Run time: {self.elapsed_time()} s"
        _logger.info(inf_msg)

    def check_exit_splitting_loop(self, k: int) -> tuple[bool, npt.NDArray[np.number]]:
        """Check for exit criterion of the splitting loop.

        Args:
            k: loop counter

        Returns:
            bool to trigger splitting loop break
            array of maximas across all trajectories
        """
        # Gather max score from all trajectories
        # and check for early convergence
        all_converged = True
        maxes = np.zeros(self._tdb.traj_list_len())
        for i in range(self._tdb.traj_list_len()):
            maxes[i] = self._tdb.get_traj(i).score_max()
            all_converged = all_converged and self._tdb.get_traj(i).is_converged()

        # Check for walltime
        if self.out_of_time():
            warn_msg = f"Ran out of time after {k} splitting iterations"
            _logger.warning(warn_msg)
            return True, maxes

        # Exit if our work is done
        if all_converged:
            inf_msg = f"All trajectories converged after {k} splitting iterations"
            _logger.info(inf_msg)
            return True, maxes

        # Exit if splitting is stalled
        if (np.amax(maxes) - np.amin(maxes)) < STALL_TOL:
            err_msg = f"Splitting is stalling with all trajectories stuck at a score_max: {np.amax(maxes)}"
            _logger.exception(err_msg)
            raise RuntimeError(err_msg)

        return False, maxes

    def finish_ongoing_splitting(self) -> None:
        """Check and finish unfinished splitting iterations.

        If the run was interrupted during a splitting iteration,
        the branched trajectories might not have ended yet. In that case,
        a list of trajectories to finish is listed in the database.
        """
        # Check the database for unfinished splitting iteration when restarting.
        # At this point, branching has been done, but advancing to final
        # time is still ongoing.
        ongoing_list = self._tdb.get_ongoing()
        if ongoing_list:
            inf_msg = f"Unfinished splitting iteration detected, traj {ongoing_list} need(s) finishing"
            _logger.info(inf_msg)
            with get_runner_type(self._parameters)(
                self._parameters, pool_worker, self._parameters.get("runner", {}).get("nworker_iter", 1)
            ) as runner:
                for i in ongoing_list:
                    t = self._tdb.get_traj(i)
                    task = [t, self._endDate, self._tdb.pool_file(), self._tdb.path()]
                    runner.make_promise(task)

                try:
                    finished_traj = runner.execute_promises()
                except Exception:
                    err_msg = f"Failed to finish branching {len(ongoing_list)} trajectories"
                    _logger.exception(err_msg)
                    raise

                _logger.info("Done with unfinished")

                for t in finished_traj:
                    self._tdb.overwrite_traj(t.id(), t)

                # Wrap up the iteration by updating its status in the
                # database and incrementing the iteration counter
                self._tdb.mark_last_splitting_iteration_as_done()

    def get_restart_at_random(self, min_idx_list: list[int]) -> list[int]:
        """Get a list of trajectory index to restart from at random.

        Select trajectories to restart from among the ones not
        in min_idx_list.

        Args:
            min_idx_list: list of trajectory index to restart from

        Returns:
            list of trajectory index to restart from
        """
        # Enable deterministic runs by setting a (different) seed
        # for each splitting iteration
        if self._parameters.get("tams", {}).get("deterministic", False):
            rng = np.random.default_rng(seed=42 * self._tdb.k_split())
        else:
            rng = np.random.default_rng()
        rest_idx = [-1] * len(min_idx_list)
        for i in range(len(min_idx_list)):
            rest_idx[i] = min_idx_list[0]
            while rest_idx[i] in min_idx_list:
                rest_idx[i] = rng.integers(low=0, high=self._tdb.traj_list_len(), dtype=int)
        return rest_idx

    def do_multilevel_splitting(self) -> None:
        """Schedule splitting of the initial ensemble of stochastic trajectories.

        Perform the multi-level splitting iterations, possibly restarting multiple
        trajectories at each iterations. All the trajectories in an iterations are
        advanced together, such that each iteration takes the maximum duration among
        the branched trajectories.

        If the walltime is exceeded, the splitting loop is stopped and ongoing
        trajectories are flagged in the database in order to finish them upon
        restart.

        The runner will use the number of workers specified in the
        input file under the runner section.

        Raises:
            Error if the runner fails
        """
        inf_msg = "Using multi-level splitting to get the probability"
        _logger.info(inf_msg)

        # Finish any unfinished splitting iteration
        self.finish_ongoing_splitting()

        # Initialize splitting iterations counter
        k = self._tdb.k_split()

        with get_runner_type(self._parameters)(
            self._parameters, ms_worker, self._parameters.get("runner", {}).get("nworker_iter", 1)
        ) as runner:
            while k < self._tdb.n_split_iter():
                inf_msg = f"Starting TAMS iter. {k} with {runner.n_workers()} workers"
                _logger.info(inf_msg)

                # Plot trajectory database scores
                if self._plot_diags:
                    pltfile = f"Score_k{k:05}.png"
                    if Path(pltfile).exists():
                        wrn_msg = f"Attempting to overwrite the plot file {pltfile}"
                        _logger.warning(wrn_msg)
                    self._tdb.plot_score_functions(pltfile)

                # Get the ensemble maximums and check for early exit conditions
                early_exit, maxes = self.check_exit_splitting_loop(k)

                # Get the nworker lower scored trajectories
                # or more if equal score
                min_idx_list, min_vals = get_min_scored(maxes, runner.n_workers())

                # Randomly select trajectory to branch from
                ancestor_idx = self.get_restart_at_random(min_idx_list)
                n_branch = len(min_idx_list)

                # Update the database with the data of the current
                # iteration
                self._tdb.append_splitting_iteration_data(
                    k, n_branch, min_idx_list, ancestor_idx, min_vals.tolist(), [np.min(maxes), np.max(maxes)]
                )

                # Query the current iteration weight
                # to compute the individual weight of each trajectory in the ensemble
                # at the end of the splitting iteration
                new_traj_weight = self._tdb.weights()[-1] / float(self._tdb.n_traj())

                # Exit the loop if needed
                if early_exit:
                    break

                # Assemble a list of promises
                # and archive the discarded trajectories
                for i in range(n_branch):
                    # Archive
                    self._tdb.archive_trajectory(self._tdb.get_traj(min_idx_list[i]))

                    # Worker task
                    task = [
                        self._tdb.get_traj(ancestor_idx[i]),
                        self._tdb.get_traj(min_idx_list[i]),
                        min_vals[i],
                        new_traj_weight,
                        self._endDate,
                        self._tdb.pool_file(),
                        self._tdb.path(),
                    ]
                    runner.make_promise(task)

                try:
                    restarted_trajs = runner.execute_promises()
                except Exception:
                    err_msg = f"Failed to branch {n_branch} trajectories at iteration {k}"
                    _logger.exception(err_msg)
                    raise

                # Update the trajectories in the database
                for t in restarted_trajs:
                    self._tdb.overwrite_traj(t.id(), t)

                # Update the weights of all trajectories with the current
                # iteration weight
                self._tdb.update_trajectories_weights()

                if self.out_of_time():
                    # Save splitting data with ongoing trajectories
                    # but do not increment splitting index yet
                    warn_msg = f"Ran out of time after {k} splitting iterations"
                    _logger.warning(warn_msg)
                    break

                # Wrap up the iteration by updating its status in the
                # database and incrementing the iteration counter
                self._tdb.mark_last_splitting_iteration_as_done()
                k = k + n_branch

    def compute_probability(self) -> float:
        """Compute the probability using TAMS.

        Returns:
            the transition probability
        """
        inf_msg = f"Computing {self._fmodel_t.name()} rare event probability using TAMS"
        _logger.info(inf_msg)

        # Generate the initial trajectory ensemble
        init_ensemble_need_work = not self._tdb.init_ensemble_done()
        if init_ensemble_need_work:
            self.generate_trajectory_ensemble()

        # Check for early convergence
        all_converged = True
        for t in self._tdb.traj_list():
            if not t.is_converged():
                all_converged = False
                break

        if init_ensemble_need_work and all_converged:
            inf_msg = "All trajectories in the ensemble converged prior to splitting !"
            _logger.info(inf_msg)
            return 1.0

        if self.out_of_time():
            warn_msg = "Ran out of walltime ! Exiting now."
            _logger.warning(warn_msg)
            return -1.0

        if self._init_ensemble_only:
            warn_msg = "Stopping after the initial ensemble stage !"
            _logger.warning(warn_msg)
            return -1.0

        # Perform multilevel splitting
        if not all_converged:
            self.do_multilevel_splitting()

        if self.out_of_time():
            warn_msg = "Ran out of walltime ! Exiting now."
            _logger.warning(warn_msg)
            return -1.0

        transition_probability = self._tdb.get_transition_probability()

        inf_msg = f"Run time: {self.elapsed_time()} s"
        _logger.info(inf_msg)

        # Load the archived trajectories data since the workers
        # discarded them but the persistent Python process did not
        # kept track
        self._tdb.load_archived_trajectories()

        self._tdb.info()

        return transition_probability

    def get_database(self) -> Database:
        """Accessor to database.

        Returns:
            A reference to the database in use
        """
        return self._tdb

    def __del__(self) -> None:
        """Destructor.

        It is mostly useful on Windows systems
        """
        # Force deletion of database
        if hasattr(self, "_tdb"):
            del self._tdb



"""A set of utility functions for TAMS."""

import ast
import inspect
import logging
import sys
import textwrap
from abc import ABCMeta
from pathlib import Path
from typing import Any
import numpy as np
import numpy.typing as npt

_logger = logging.getLogger(__name__)

def is_windows_os() -> bool:
    """Indicates Windows platform."""
    system = sys.platform.lower()
    return system.startswith("win")

def is_mac_os() -> bool:
    """Indicates MacOS platform."""
    system = sys.platform.lower()
    return system.startswith("dar")

def setup_logger(params: dict[Any, Any]) -> None:
    """Setup the logger parameters.

    Args:
        params: a dictionary of parameters
    """
    # Set logging level
    log_level_str = params["tams"].get("loglevel", "INFO")
    if log_level_str.upper() == "DEBUG":
        log_level = logging.DEBUG
    elif log_level_str.upper() == "INFO":
        log_level = logging.INFO
    elif log_level_str.upper() == "WARNING":
        log_level = logging.WARNING
    elif log_level_str.upper() == "ERROR":
        log_level = logging.ERROR

    log_format = "[%(levelname)s] %(asctime)s - %(message)s"

    # Set root logger
    logging.basicConfig(
        level=log_level,
        format=log_format,
    )

    # Add file handler to root logger
    if params["tams"].get("logfile", None):
        log_file = logging.FileHandler(params["tams"]["logfile"])
        log_file.setLevel(log_level)
        log_file.setFormatter(logging.Formatter(log_format))
        logging.getLogger("").addHandler(log_file)


def get_min_scored(maxes: npt.NDArray[Any], nworkers: int) -> tuple[list[int], npt.NDArray[Any]]:
    """Get the nworker lower scored trajectories or more if equal score.

    Args:
        maxes: array of maximas across all trajectories
        nworkers: number of workers

    Returns:
        list of indices of the nworker lower scored trajectories
        array of minimas
    """
    ordered_tlist = np.argsort(maxes)
    is_same_min = False
    min_idx_list: list[int] = []
    for idx in ordered_tlist:
        if len(min_idx_list) > 0:
            is_same_min = maxes[idx] == maxes[min_idx_list[-1]]
        if len(min_idx_list) < nworkers or is_same_min:
            min_idx_list.append(int(idx))

    min_vals = maxes[min_idx_list]
    return min_idx_list, min_vals


def moving_avg(arr_in: npt.NDArray[Any], window_l: int) -> npt.NDArray[Any]:
    """Return the moving average of a 1D numpy array.

    Args:
        arr_in: 1D numpy array
        window_l: length of the moving average window

    Returns:
        1D numpy array
    """
    arr_out = np.zeros(arr_in.shape[0])
    for i in range(len(arr_in)):
        lbnd = max(i - int(np.ceil(window_l / 2)), 0)
        hbnd = min(i + int(np.floor(window_l / 2)), len(arr_in) - 1)
        if lbnd == 0:
            hbnd = window_l
        if hbnd == len(arr_in) - 1:
            lbnd = len(arr_in) - window_l - 1
        arr_out[i] = np.mean(arr_in[lbnd:hbnd])
    return arr_out


def get_module_local_import(module_name: str) -> list[str]:
    """Helper function getting local imported mods list.

    When pickling the forward model code, the model itself can import from
    several other local files. We also want to pickle those by value so let's get
    the list.

    Args:
        module_name: a module name we want the locally imported modules

    Returns:
        A list of local modules names imported within the provide module
    """
    # Check that module exists
    if module_name not in sys.modules:
        err_msg = f"Attempting to extract sub import from {module_name} missing from currently loaded modules"
        _logger.exception(err_msg)
        raise ValueError(err_msg)

    # Check access to the module file
    if hasattr(sys.modules[module_name], "__file__") and Path(str(sys.modules[module_name].__file__)).exists():
        mfile = Path(str(sys.modules[module_name].__file__))
    else:
        err_msg = f"Attempting to locate sub import file from {module_name}, but file is missing or undefined"
        _logger.exception(err_msg)
        raise FileNotFoundError(err_msg)

    # Parse the module file
    # for imports
    with mfile.open("r") as f:
        file_raw = f.read()

    file_ast = ast.parse(file_raw)
    all_modules = []

    for node in ast.walk(file_ast):
        # Append "import X" type
        if isinstance(node, ast.Import):
            all_modules.extend([x.name for x in node.names])
        # Append "from X import Y" type
        if isinstance(node, ast.ImportFrom) and node.module:
            all_modules.append(node.module)

    # Return only those whose file is in the current folder
    # or from the 'examples' folder of pyTAMS
    return [
        m
        for m in all_modules
        if (
            hasattr(sys.modules[m], "__file__")
            and (
                Path(str(sys.modules[m].__file__)).parent == Path().absolute()
                or any((p.name == "examples") for p in Path(str(sys.modules[m].__file__)).parents)
            )
        )
    ]


def generate_subclass(abc_cls: ABCMeta, class_name: str, file_path: str) -> None:
    """Generate a subclass skeleton.

    Implementing all abstract methods from `abc_cls`, written to `file_path`.

    Args:
        abc_cls: an ABC
        class_name: the new subclass name
        file_path: where to write the subclass
    """
    # Identify abstract methods
    abstract_methods = {
        name: value for name, value in abc_cls.__dict__.items() if getattr(value, "__isabstractmethod__", False)
    }

    # Build import line
    module_name = abc_cls.__module__
    abc_name = abc_cls.__name__
    import_lines = [f"from {module_name} import {abc_name}\n", "import typing\n", "from typing import Any\n\n\n"]

    # Build class header
    lines = [*import_lines, f"class {class_name}({abc_name}):\n"]
    lines.append('    """TODO: add class docstring."""\n')

    if not abstract_methods:
        lines.append("    pass\n")
    else:
        # Generate each required method with preserved signature
        for name, func in abstract_methods.items():
            sig = inspect.signature(func)
            doc = inspect.getdoc(func)

            lines.append(f"    def {name}{sig}:\n")
            if doc:
                # Indent docstring correctly
                doc_clean = textwrap.indent('"""' + doc + '\n"""', " " * 8)
                lines.append(f"{doc_clean}\n")
            else:
                lines.append('        """TODO: implement method."""\n')

            lines.append("        # Implement concrete method body\n\n")

    # Add the name class method
    lines.append("    @classmethod\n")
    lines.append("    def name(cls) -> str:\n")
    lines.append('        """Return a the model name."""\n')
    lines.append(f'        return "{class_name}"\n')

    # Write to file
    Path(file_path).write_text("".join(lines))



"""Tests for the pytams.sqlpool class."""
from pathlib import Path
import numpy as np
import pytest
from sqlalchemy.exc import SQLAlchemyError
from pytams.sqldb import SQLFile


def test_createdb():
    """Initialize a SQLFile."""
    poolfile = SQLFile("test.db")
    assert poolfile.name() == "test.db"
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_createdb_inmemory():
    """Initialize a SQLFile in memory."""
    poolfile = SQLFile("", in_memory=True)
    assert poolfile.name() == ""

def test_createdb_read_only():
    """Initialize a read only SQLFile."""
    with pytest.raises(SQLAlchemyError):
        _ = SQLFile("testRO.db", ro_mode=True)

@pytest.mark.usefixtures("skip_on_windows")
def test_createdb_fail():
    """Fail to initialize a SQLFile."""
    with pytest.raises(SQLAlchemyError):
        _ = SQLFile("/test.db")

def test_add_traj_to_db():
    """Add a trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    assert poolfile.get_trajectory_count() == 1
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

@pytest.mark.usefixtures("skip_on_windows")
def test_add_traj_to_ro_db():
    """Try add a trajectory to an RO SQLFile."""
    poolfile = SQLFile("test.db") # First create the DB
    poolfile = SQLFile("test.db", ro_mode=True) # Open in RO
    with pytest.raises(SQLAlchemyError):
        poolfile.add_trajectory("test.xml","")
    Path("./test.db").unlink(missing_ok=True)

def test_add_traj_and_update_to_db():
    """Add and update a trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    assert poolfile.fetch_trajectory(0)[0] == "test.xml"
    poolfile.update_trajectory(0, "UpdatedTest.xml", "")
    assert poolfile.fetch_trajectory(0)[0] == "UpdatedTest.xml"
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

@pytest.mark.usefixtures("skip_on_windows")
def test_try_update_traj_to_db():
    """Try update missing trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    with pytest.raises(SQLAlchemyError):
        poolfile.update_trajectory(0, "UpdatedTest.xml", "dummy")
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

@pytest.mark.usefixtures("skip_on_windows")
def test_try_update_weight_to_db():
    """Try updating weight to missing trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    with pytest.raises(SQLAlchemyError):
        poolfile.update_trajectory_weight(3, 1.0)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_add_traj_to_db_inmemory():
    """Add a trajectory to SQL database in memory."""
    poolfile = SQLFile("", in_memory=True)
    poolfile.add_trajectory("test.xml","")
    assert poolfile.get_trajectory_count() == 1

@pytest.mark.usefixtures("skip_on_windows")
def test_add_traj_to_missing_db():
    """Add a trajectory to a deleted SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    assert poolfile.get_trajectory_count() == 1
    Path("./test.db").unlink(missing_ok=True)
    with pytest.raises(SQLAlchemyError):
        poolfile.add_trajectory("test2.xml","")

def test_archive_traj_to_db():
    """Archive a trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.archive_trajectory("test.xml","")
    assert poolfile.get_archived_trajectory_count() == 1
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_add_traj_and_fetch():
    """Add a trajectory and fetch from SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    poolfile.add_trajectory("test_2.xml","")
    assert poolfile.get_trajectory_count() == 2
    traj, metadata = poolfile.fetch_trajectory(0)
    assert traj == "test.xml"
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_fetch_unknown_traj():
    """Fetch an unknown trajectory."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    assert poolfile.get_trajectory_count() == 1
    with pytest.raises(ValueError):
        _, _ = poolfile.fetch_trajectory(1)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_archive_and_fetch_traj_to_db():
    """Archive a trajectory to SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.archive_trajectory("test.xml","")
    assert poolfile.get_archived_trajectory_count() == 1
    traj, metadata = poolfile.fetch_archived_trajectory(0)
    assert traj == "test.xml"
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_fetch_unknown_archived_traj():
    """Fetch an unknown archived trajectory."""
    poolfile = SQLFile("test.db")
    poolfile.archive_trajectory("test.xml","")
    assert poolfile.get_archived_trajectory_count() == 1
    with pytest.raises(ValueError):
        _, _ = poolfile.fetch_archived_trajectory(1)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_trajectory():
    """Lock a trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    status = poolfile.lock_trajectory(0)
    assert status
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_locked_trajectory():
    """Lock an already locked trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    status = poolfile.lock_trajectory(0)
    status = poolfile.lock_trajectory(0)
    assert status is False
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_and_release_trajectory():
    """Lock and release a trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    status = poolfile.lock_trajectory(0)
    poolfile.release_trajectory(0)
    status = poolfile.lock_trajectory(0)
    assert status
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_and_complete_trajectory():
    """Lock and mark complete a trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    _ = poolfile.lock_trajectory(0)
    poolfile.mark_trajectory_as_completed(0)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_and_complete_unknown_trajectory():
    """Lock and try to mark a trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    _ = poolfile.lock_trajectory(0)
    with pytest.raises(ValueError):
        poolfile.mark_trajectory_as_completed(1)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_and_release_multiple_trajectory():
    """Lock and release several trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    for i in range(10):
        poolfile.add_trajectory(f"test{i}.xml","")
        status = poolfile.lock_trajectory(0)
    poolfile.release_all_trajectories()
    status = True
    for _ in range(10):
        status = status or poolfile.lock_trajectory(0)
    assert status
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_lock_unknown_trajectory():
    """Lock an unknown trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    with pytest.raises(ValueError):
        _ = poolfile.lock_trajectory(1)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

@pytest.mark.usefixtures("skip_on_windows")
def test_lock_in_missing_db():
    """Lock a trajectory in a missing SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    Path("./test.db").unlink(missing_ok=True)
    with pytest.raises(SQLAlchemyError):
        _ = poolfile.lock_trajectory(0)

def test_release_unknown_trajectory():
    """Release an unknown trajectory in the SQLFile."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    with pytest.raises(ValueError):
        poolfile.release_trajectory(1)
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_splitting_data_add():
    """Adding splitting data to the database."""
    poolfile = SQLFile("test.db")
    for i in range(10):
        poolfile.add_splitting_data(i, 1, 0.1, [i-1], [0], [0.0], [0.0, 0.0])
        poolfile.mark_last_iteration_as_completed()
    del poolfile
    Path("./test.db").unlink(missing_ok=True)

def test_splitting_data_add_and_ongoing():
    """Adding splitting data to the database."""
    poolfile = SQLFile("", in_memory=True)
    poolfile.mark_last_iteration_as_completed()
    for i in range(10):
        poolfile.add_splitting_data(i, 1, 0.1, [i-1], [0], [0.0], [0.0, 0.0])
        poolfile.mark_last_iteration_as_completed()
    assert poolfile.get_ongoing() is None
    poolfile.add_splitting_data(10, 1, 0.1, [10-1,1,56], [0], [0.0], [0.0, 0.0])
    assert poolfile.get_ongoing() == [9,1,56]

def test_splitting_data_add_and_query():
    """Adding splitting data to the database."""
    poolfile = SQLFile("", in_memory=True)
    for i in range(1,2):
        poolfile.add_splitting_data(2*i, 1, 0.1, [2*i-1], [0], [0.0], [0.0, 0.0])
        poolfile.mark_last_iteration_as_completed()
    assert np.all(poolfile.get_minmax()[0] == np.array([2.0,0.0,0.0], dtype="float32"))

def test_splitting_data_add_update_and_query():
    """Adding splitting data to the database."""
    poolfile = SQLFile("", in_memory=True)
    poolfile.add_splitting_data(2, 1, 0.1, [1], [0], [0.0], [0.0, 0.0])
    poolfile.update_splitting_data(2, 1, 0.1, [1], [0], [0.0], [0.0, 0.3])
    poolfile.mark_last_iteration_as_completed()
    assert np.all(poolfile.get_minmax()[0] == np.array([2.0,0.0,0.3], dtype="float32"))

@pytest.mark.usefixtures("skip_on_windows")
def test_splitting_data_query_fail():
    """Adding splitting data to the database."""
    poolfile = SQLFile("test.db")
    for i in range(1):
        poolfile.add_splitting_data(2*i, 1, 0.1, [2*i-1], [0], [0.0], [0.0, 0.0])
    assert poolfile.get_k_split() == 1

    poolfile = SQLFile("test.db", ro_mode=True)
    with pytest.raises(SQLAlchemyError):
        poolfile.mark_last_iteration_as_completed()
    Path("./test.db").unlink(missing_ok=True)

def test_dump_json():
    """Dump the content of the DB to a json file."""
    poolfile = SQLFile("test.db")
    poolfile.add_trajectory("test.xml","")
    poolfile.archive_trajectory("test_arch.xml","")
    poolfile.dump_file_json()
    assert Path("./test.json").exists() is True
    del poolfile
    Path("./test.db").unlink(missing_ok=True)
    Path("./test.json").unlink(missing_ok=True)



"""Tests for the pytams.tams class."""

import logging
import shutil
from pathlib import Path
import pytest
import toml
from pytams.database import Database
from pytams.tams import TAMS
from pytams.utils import is_mac_os
from tests.models import DoubleWellModel
from tests.models import SimpleFModel


def test_init_tams():
    """Test TAMS initialization."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {"tams": {"ntrajectories": 500, "nsplititer": 200}, "trajectory": {"end_time": 0.02, "step_size": 0.001}}, f
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    assert tams.n_traj() == 500
    Path("input.toml").unlink(missing_ok=True)


def test_init_tams_missing_req():
    """Test failed TAMS initialization."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump({"tams": {"nsplititer": 200}, "trajectory": {"end_time": 0.02, "step_size": 0.001}}, f)
    with pytest.raises(ValueError):
        _ = TAMS(fmodel_t=fmodel, a_args=[])
    Path("input.toml").unlink(missing_ok=True)


def test_init_tams_no_input():
    """Test failed TAMS initialization."""
    fmodel = SimpleFModel
    with pytest.raises(ValueError):
        _ = TAMS(fmodel_t=fmodel, a_args=["-i", "dummy.toml"])


def test_simple_model_tams():
    """Test TAMS with simple model."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "WARNING"},
                "runner": {"type": "asyncio"},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 0.15},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba == 1.0
    Path("input.toml").unlink(missing_ok=True)


def test_simple_model_init_ensemble_stage_tams(caplog: pytest.LogCaptureFixture):
    """Test TAMS with simple model."""
    caplog.set_level(logging.WARNING)
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "WARNING", "init_ensemble_only": True},
                "runner": {"type": "asyncio"},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 1.15},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    assert "Stopping after the initial ensemble stage !" in caplog.text
    Path("input.toml").unlink(missing_ok=True)


def test_simple_model_init_ensemble_stage_and_continue_tams():
    """Test TAMS with simple model."""
    fmodel = SimpleFModel
    params_dict = {
        "tams": {"ntrajectories": 10, "nsplititer": 200, "loglevel": "WARNING", "init_ensemble_only": True},
        "runner": {"type": "asyncio"},
        "database": {"path": "simpleModelTest.tdb"},
        "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 1.15},
    }
    with Path("input.toml").open("w") as f:
        toml.dump(params_dict, f)
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    del tams
    tdb = Database.load(Path("simpleModelTest.tdb"))
    assert tdb.n_traj() == 10
    del tdb
    params_dict["tams"]["ntrajectories"] = 20
    with Path("input.toml").open("w") as f:
        toml.dump(params_dict, f)
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    del tams
    tdb = Database.load(Path("simpleModelTest.tdb"))
    assert tdb.n_traj() == 20
    del tdb
    Path("input.toml").unlink(missing_ok=True)
    shutil.rmtree("simpleModelTest.tdb")


def test_simple_model_tams_with_db():
    """Test TAMS with simple model."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "WARNING"},
                "runner": {"type": "dask"},
                "database": {"path": "simpleModelTest.tdb"},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 0.15, "chkfile_dump_all": True},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    del tams
    assert transition_proba == 1.0
    shutil.rmtree("simpleModelTest.tdb")
    Path("input.toml").unlink(missing_ok=True)

def test_simple_model_tams_with_db_access():
    """Test TAMS with simple model and access to the database."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "WARNING"},
                "runner": {"type": "dask"},
                "database": {"path": "simpleModelTest.tdb"},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 0.15, "chkfile_dump_all": True},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    tdb = tams.get_database()
    assert tdb.get_transition_probability() == 1
    del tams
    del tdb
    shutil.rmtree("simpleModelTest.tdb")
    Path("input.toml").unlink(missing_ok=True)

def test_simple_model_tams_slurm_fail():
    """Test TAMS with simple model with Slurm dask backend."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "DEBUG"},
                "runner": {"type": "dask"},
                "dask": {"backend": "slurm", "slurm_config_file": "dummy.yaml"},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 0.15},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    with pytest.raises(FileNotFoundError):
        tams.compute_probability()
    Path("input.toml").unlink(missing_ok=True)


@pytest.mark.usefixtures("skip_on_windows")
def test_simple_model_twice_tams():
    """Test TAMS with simple model."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "WARNING", "logfile": "test.log"},
                "runner": {"type": "asyncio"},
                "database": {"path": "simpleModelTest.tdb", "restart": True},
                "trajectory": {"end_time": 0.02, "step_size": 0.001, "targetscore": 0.15},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba == 1.0
    del tams
    # Re-init TAMS and run to test competing database
    # on disk.
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    del tams
    ndb = 0
    for folder in Path("./").iterdir():
        if "simpleModelTest" in str(folder):
            shutil.rmtree(folder)
            ndb += 1
    assert ndb == 2
    assert Path("test.log").exists()
    Path("test.log").unlink(missing_ok=True)
    Path("input.toml").unlink(missing_ok=True)


def test_stalling_simplemodel_tams():
    """Test TAMS with simple model and stalled score function."""
    fmodel = SimpleFModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 200, "loglevel": "ERROR"},
                "runner": {"type": "asyncio"},
                "trajectory": {"end_time": 1.0, "step_size": 0.01, "targetscore": 1.1},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    with pytest.raises(RuntimeError):
        tams.compute_probability()


def test_doublewell_tams():
    """Test TAMS with the doublewell model."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 50, "nsplititer": 200, "walltime": 500.0},
                "runner": {"type": "dask"},
                "model": {"noise_amplitude": 0.8},
                "trajectory": {"end_time": 6.0, "step_size": 0.01, "targetscore": 0.8},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba >= 0.2
    Path("input.toml").unlink(missing_ok=True)


def test_doublewell_save_tams():
    """Test TAMS with the doublewell model."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 50, "nsplititer": 100, "walltime": 500.0},
                "runner": {"type": "dask"},
                "database": {"path": "dwTest.tdb"},
                "model": {"noise_amplitude": 0.8},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.3},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba >= 0.2
    del tams
    Path("input.toml").unlink(missing_ok=True)
    shutil.rmtree("dwTest.tdb")


def test_doublewell_deterministic_tams():
    """Test TAMS with the doublewell model."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 400, "walltime": 500.0, "deterministic": True},
                "runner": {"type": "asyncio"},
                "model": {"noise_amplitude": 0.8},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.8},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba == 0.5471567691116268
    Path("input.toml").unlink(missing_ok=True)


@pytest.mark.usefixtures("skip_on_windows")
def test_doublewell_deterministic_tams_with_diags(caplog: pytest.LogCaptureFixture):
    """Test TAMS with the doublewell model."""
    caplog.set_level(logging.WARNING)
    fmodel = DoubleWellModel
    Path("Score_k00001.png").touch()
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {
                    "ntrajectories": 5,
                    "nsplititer": 5,
                    "walltime": 500.0,
                    "deterministic": True,
                    "diagnostics": True,
                },
                "runner": {"type": "asyncio"},
                "model": {"noise_amplitude": 0.4},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.8},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    assert "Attempting to overwrite the plot file" in caplog.text
    Path("input.toml").unlink(missing_ok=True)
    for p in Path().glob("Score*.png"):
        p.unlink()


@pytest.mark.dependency
def test_doublewell_2_workers_tams():
    """Test TAMS with the doublewell model using two workers."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 50, "nsplititer": 400, "walltime": 500.0, "deterministic": True},
                "runner": {"type": "dask", "nworker_init": 2, "nworker_iter": 2},
                "model": {"noise_amplitude": 0.8},
                "database": {"path": "dwTest.tdb", "archive_discarded": True},
                "trajectory": {"end_time": 5.0, "step_size": 0.01, "targetscore": 0.5},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba == 0.692533980184018
    del tams
    Path("input.toml").unlink(missing_ok=True)


@pytest.mark.dependency(depends=["test_doublewell_2_workers_tams"])
def test_doublewell_2_workers_load_db():
    """Load the database from previous test."""
    tdb = Database.load(Path("dwTest.tdb"))
    tdb.load_data(True)
    assert tdb.traj_list_len() == 50
    assert tdb.archived_traj_list_len() == 16
    del tdb


@pytest.mark.dependency(depends=["test_doublewell_2_workers_tams"])
def test_doublewell_2_workers_restore_tams():
    """Test TAMS with the doublewell model using two workers and restoring."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 100, "nsplititer": 400, "walltime": 500.0},
                "database": {"path": "dwTest.tdb"},
                "runner": {"type": "asyncio", "nworker_init": 2, "nworker_iter": 2},
                "model": {"noise_amplitude": 0.8},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.6},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba >= 0.2
    Path("input.toml").unlink(missing_ok=True)
    del tams
    shutil.rmtree("dwTest.tdb")


def test_doublewell_very_slow_tams():
    """Test TAMS run out of time with a slow doublewell."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 10, "nsplititer": 400, "walltime": 3.0},
                "database": {"path": "vslowdwTest.tdb"},
                "runner": {"type": "dask", "nworker_init": 1, "nworker_iter": 1},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.7},
                "model": {"slow_factor": 0.01, "noise_amplitude": 0.1},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba <= 0.0
    Path("input.toml").unlink(missing_ok=True)
    del tams
    shutil.rmtree("vslowdwTest.tdb")


@pytest.mark.dependency
def test_doublewell_slow_tams_stop():
    """Test TAMS run out of time with a slow doublewell."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 8, "nsplititer": 400, "walltime": 3.0},
                "database": {"path": "slowdwTest.tdb"},
                "runner": {"type": "asyncio", "nworker_init": 1, "nworker_iter": 1},
                "trajectory": {"end_time": 8.0, "step_size": 0.01, "targetscore": 0.9},
                "model": {"slow_factor": 0.0002, "noise_amplitude": 0.1},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba <= 0.0
    del tams
    Path("input.toml").unlink(missing_ok=True)


@pytest.mark.dependency(depends=["test_doublewell_slow_tams_stop"])
def test_doublewell_slow_tams_restore_during_initial_ensemble():
    """Test TAMS restarting a slow doublewell."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 8, "nsplititer": 400, "walltime": 8.0},
                "database": {"path": "slowdwTest.tdb"},
                "runner": {"type": "asyncio", "nworker_init": 1, "nworker_iter": 1},
                "trajectory": {"end_time": 8.0, "step_size": 0.01, "targetscore": 0.9},
                "model": {"slow_factor": 0.0002, "noise_amplitude": 0.1},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    assert transition_proba <= 0.0
    del tams
    Path("input.toml").unlink(missing_ok=True)


@pytest.mark.usefixtures("skip_on_windows")
@pytest.mark.dependency(depends=["test_doublewell_slow_tams_restore_during_initial_ensemble"])
def test_doublewell_slow_tams_restore_during_splitting(caplog: pytest.LogCaptureFixture):
    """Test TAMS restarting a slow doublewell."""
    caplog.set_level(logging.INFO)
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 8, "nsplititer": 400, "walltime": 2.0},
                "database": {"path": "slowdwTest.tdb"},
                "runner": {"type": "asyncio", "nworker_init": 1, "nworker_iter": 1},
                "trajectory": {"end_time": 8.0, "step_size": 0.01, "targetscore": 0.9},
                "model": {"slow_factor": 0.0002, "noise_amplitude": 0.1},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    _ = tams.compute_probability()
    assert "Unfinished splitting iteration detected" in caplog.text
    Path("input.toml").unlink(missing_ok=True)
    del tams
    shutil.rmtree("slowdwTest.tdb")


@pytest.mark.usefixtures("skip_on_windows")
def test_doublewell_slow_tams_restore_more_split():
    """Test restart TAMS more splitting iterations."""
    fmodel = DoubleWellModel
    params_dict = {
        "tams": {"ntrajectories": 20, "nsplititer": 20, "walltime": 20.0, "deterministic": True},
        "database": {"path": "dwTest.tdb"},
        "runner": {"type": "asyncio", "nworker_init": 2, "nworker_iter": 1},
        "trajectory": {"end_time": 6.0, "step_size": 0.01, "targetscore": 0.6},
        "model": {"slow_factor": 0.00000001, "noise_amplitude": 0.6},
    }
    with Path("input.toml").open("w") as f:
        toml.dump(params_dict, f)
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams.compute_probability()
    del tams
    assert transition_proba == 0.08937321391676148
    params_dict["tams"]["nsplititer"] = 30
    with Path("input.toml").open("w") as f:
        toml.dump(params_dict, f)
    tams_load = TAMS(fmodel_t=fmodel, a_args=[])
    transition_proba = tams_load.compute_probability()
    # Not sure why this particular test is platform dependent
    if is_mac_os():
        assert transition_proba == 0.0853804929982172
    else:
        assert transition_proba == 0.07470793360861466
    Path("input.toml").unlink(missing_ok=True)
    del tams_load
    shutil.rmtree("dwTest.tdb")



from __future__ import annotations
import json
import logging
import shutil
import time
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import cast
import numpy as np
import numpy.typing as npt
from pytams.xmlutils import dict_to_xml
from pytams.xmlutils import make_xml_snapshot
from pytams.xmlutils import read_xml_snapshot

if TYPE_CHECKING:
    from pytams.fmodel import ForwardModelBaseClass

_logger = logging.getLogger(__name__)


class WallTimeLimitError(Exception):
    """Exception for running into wall time limit."""


def form_trajectory_id(n: int, nb: int = 0) -> str:
    """Helper to assemble a trajectory ID string.

    Args:
        n : trajectory index
        nb : number of branching

    Returns:
        trajectory ID
    """
    return f"traj{n:06}_{nb:04}"


def get_index_from_id(identity: str) -> tuple[int, int]:
    """Helper to get trajectory index from ID string.

    Args:
        identity : trajectory ID

    Returns:
        trajectory index and number of branching
    """
    return int(identity[-10:-5]), int(identity[-4:])


@dataclass
class Snapshot:
    """A dataclass defining a snapshot.

    Gathering what defines a snapshot into an object.
    The time and score are of float type, but the
    actual type of the noise and state are completely
    determined by the forward model.
    A snapshot is allowed to have a state or not to
    accommodate memory savings.

    Attributes:
        time : snapshot time
        score : score function value
        noise : noise used to reach this snapshot
        state : model state
    """

    time: float
    score: float
    noise: Any
    state: Any | None = None

    def has_state(self) -> bool:
        """Check if snapshot has state.

        Returns:
            bool : True if state is not None
        """
        return self.state is not None


class Trajectory:
    """A class defining a stochastic trajectory.

    The trajectory class is a container for time-ordered snapshots.
    It contains an instance of the forward model, current and end times, and
    a list of the model snapshots. Note that the class uses a plain list of snapshots
    and not a more computationally efficient data structure such as a numpy array
    for convenience. It is assumed that the computational cost of running TAMS
    reside in the forward model and the overhead of the trajectory class is negligible.

    It also provide the forward model with the necessary context to advance in time,
    method to move forward in time, methods to save/load the trajectory to/from disk
    as well as accessor to the trajectory history (time, state, score, ...).

    The _computed_steps variable store the number of steps actually taken by the
    trajectory. It differs from the _step variable when a trajectory is branched
    from an ancestor.

    Attributes:
        _parameters_full : the full parameters dictionary
        _tid : the trajectory index
        _checkFile : the trajectory checkpoint file
        _workdir : the model working directory
        _score_max : the maximum score
        _snaps : a list of snapshots
        _step : the current step counter
        _computed_steps : the number of steps explicitly advanced by the trajectory
        _t_cur : the current time
        _t_end : the end time
        _dt : the stochastic time step size
    """

    def __init__(
        self,
        traj_id: int,
        weight: float,
        fmodel_t: type[ForwardModelBaseClass] | None,
        parameters: dict[Any, Any],
        workdir: Path | None = None,
        frozen: bool = False,
    ) -> None:
        """Create a trajectory.

        Args:
            traj_id: a int for the trajectory index
            weight: the trajectory weight in the ensemble
            fmodel_t: the forward model type
            parameters: a dictionary of input parameters
            workdir: an optional working directory
            frozen: whether the trajectory is frozen (no fmodel)
        """
        # Stash away the full parameters dict
        self._parameters_full: dict[Any, Any] = parameters

        traj_params = parameters.get("trajectory", {})
        if "end_time" not in traj_params or "step_size" not in traj_params:
            err_msg = "Trajectory 'end_time' and 'step_size' must be specified in the input file !"
            _logger.error(err_msg)
            raise ValueError

        # The workdir is a runtime parameter, not saved in the chkfile.
        self._tid: int = traj_id
        self._workdir: Path = Path.cwd() if workdir is None else workdir
        self._score_max: float = -1.0e12
        self._has_ended: bool = False
        self._has_converged: bool = False
        self._computed_steps: int = 0
        self._weight: float = weight

        # TAMS is expected to start at t = 0.0, but the forward model
        # itself can have a different internal starting point
        # or an entirely different time scale.
        self._step: int = 0
        self._t_cur: float = 0.0
        self._t_end: float = traj_params.get("end_time")
        self._dt: float = traj_params.get("step_size")

        # Trajectory convergence is defined by a target score, with
        # the score provided by the forward model, mapping the model state to
        # a s \in [0,1]. A default value of 0.95 is provided.
        self._convergedVal: float = traj_params.get("targetscore", 0.95)

        # List of snapshots
        self._snaps: list[Snapshot] = []

        # When using sparse state or for other reasons, the noise for the next few
        # steps might be already available. This backlog is used to store them.
        self.noise_backlog: list[Any] = []

        # Keep track of the branching history during TAMS
        # iterations
        self._branching_history: list[int] = []

        # For large models, the state may not be available at each snapshot due
        # to memory constraint (both in-memory and on-disk). Sparse state can
        # be specified. Finally, writing a chkfile to disk at each step might
        # incur a performance hit and is by default disabled.
        self._sparse_state_int: int = traj_params.get("sparse_freq", 1)
        self._sparse_state_beg: int = traj_params.get("sparse_start", 0) + 1
        self._write_chkfile_all: bool = traj_params.get("chkfile_dump_all", False)
        self._checkFile: Path = Path(f"{self.idstr()}.xml")

        # Each trajectory has its own instance of the forward model
        if frozen or fmodel_t is None:
            self._fmodel = None
        else:
            self._fmodel = fmodel_t(self._tid * 10000 + self.get_nbranching(), parameters, self._workdir)

    def set_checkfile(self, path: Path) -> None:
        """Setter of the trajectory checkFile.

        Args:
            path: the new checkFile
        """
        self._checkFile = path

    def set_workdir(self, path: Path) -> None:
        """Setter of the trajectory working directory.

        And propagate the workdir to the forward model.

        Args:
            path: the new working directory
        """
        self._workdir = path
        if self._fmodel is not None:
            self._fmodel.set_workdir(path)

    def get_workdir(self) -> Path:
        """Get the trajectory working directory.

        Returns:
            the working directory
        """
        return self._workdir

    def id(self) -> int:
        """Return trajectory Id.

        Returns:
            the trajectory id
        """
        return self._tid

    def idstr(self) -> str:
        """Return trajectory Id as a padded string.

        Returns:
            the trajectory id as a string
        """
        return form_trajectory_id(self._tid, self.get_nbranching())

    def advance(self, t_end: float = 1.0e12, walltime: float = 1.0e12) -> None:
        """Advance the trajectory to a prescribed end time.

        This is the main time loop of the trajectory object.
        Unless specified otherwise, the trajectory will advance until
        the end time is reached or the model has converged.

        If the walltime limit is reached, a WallTimeLimitError exception is raised.
        Note that this exception is treated as a warning not an error by the
        TAMS workers.

        Args:
            t_end: the end time of the advance
            walltime: a walltime limit to advance the model to t_end

        Returns:
            None

        Raises:
            WallTimeLimitError: if the walltime limit is reached
            RuntimeError: if the model advance run into a problem
        """
        start_time = time.monotonic()
        remaining_time = walltime - time.monotonic() + start_time
        end_time = min(t_end, self._t_end)

        if not self._fmodel:
            err_msg = f"Trajectory {self.idstr()} is frozen, without forward model. Advance() deactivated."
            _logger.error(err_msg)
            raise RuntimeError(err_msg)

        while self._t_cur < end_time and not self._has_converged and remaining_time >= 0.05 * walltime:
            # Do a single step and keep track of remaining walltime
            _ = self._one_step()

            remaining_time = walltime - time.monotonic() + start_time

        if self._t_cur >= self._t_end or self._has_converged:
            self._has_ended = True

        if self._has_ended:
            self._fmodel.clear()

        if remaining_time < 0.05 * walltime:
            warn_msg = f"{self.idstr()} ran out of time in advance()"
            _logger.warning(warn_msg)
            raise WallTimeLimitError(warn_msg)

    def _one_step(self) -> float:
        """Perform a single step of the forward model.

        Perform a single time step of the forward model. This
        function will also set the noise to use for the next step
        in the forward model if a backlog is available.
        """
        if not self._fmodel:
            err_msg = f"Trajectory {self.idstr()} is frozen, without forward model. Advance() deactivated."
            _logger.exception(err_msg)
            raise RuntimeError(err_msg)

        # Add the initial snapshot to the list
        if self._step == 0:
            self.setup_noise()
            self._append_snapshot()

        # Trigger storing the end state of the current time step
        # if the next trajectory snapshot needs it
        need_end_state = (self._sparse_state_beg + self._step + 1) % self._sparse_state_int == 0

        try:
            dt = self._fmodel.advance(self._dt, need_end_state)
        except Exception:
            err_msg = f"ForwardModel advance error at step {self._step:08}"
            _logger.exception(err_msg)
            raise

        self._step += 1
        self._t_cur = self._t_cur + dt
        score = self._fmodel.score()

        # Prepare the noise for the next step
        self.setup_noise()

        # Append a snapshot at the beginning of the time step
        self._append_snapshot(score)

        if self._write_chkfile_all:
            self.store()

        self._score_max = max(self._score_max, score)

        # The default ABC method simply check for a score above
        # the target value, but concrete implementations can override
        # with mode complex convergence criteria
        self._has_converged = self._fmodel.check_convergence(self._step, self._t_cur, score, self._convergedVal)

        # Increment the computed step counter
        self._computed_steps += 1

        return score

    def setup_noise(self) -> None:
        """Prepare the noise for the next step."""
        # Set the noise for the next model step
        # if a noise backlog is available, use it otherwise
        # make a new noise increment
        if self._fmodel:
            if self.noise_backlog:
                self._fmodel.set_noise(self.noise_backlog.pop())
            else:
                self._fmodel.set_noise(self._fmodel.make_noise())

    def _append_snapshot(self, score: float | None = None) -> None:
        """Append the current snapshot to the trajectory list."""
        # Append the current snapshot to the trajectory list
        if self._fmodel:
            need_state = (self._sparse_state_beg + self._step) % self._sparse_state_int == 0 or self._step == 0
            self._snaps.append(
                Snapshot(
                    time=self._t_cur,
                    score=score if score else self._fmodel.score(),
                    noise=self._fmodel.get_noise(),
                    state=self._fmodel.get_current_state() if need_state else None,
                ),
            )

    @classmethod
    def init_from_metadata(
        cls,
        metadata_json: str,
        fmodel_t: type[ForwardModelBaseClass],
        parameters: dict[Any, Any],
        workdir: Path | None = None,
        frozen: bool = False,
    ) -> Trajectory:
        """Initialize a trajectory from serialized metadata."""
        metadata = cls.deserialize_metadata(metadata_json)

        traj = Trajectory(
            traj_id=metadata["id"],
            weight=metadata["weight"],
            fmodel_t=fmodel_t,
            parameters=parameters,
            workdir=workdir,
            frozen=frozen,
        )

        traj._t_end = metadata["t_end"]
        traj._t_cur = metadata["t_cur"]
        traj._dt = metadata["dt"]
        traj._score_max = metadata["score_max"]
        traj._has_ended = metadata["ended"]
        traj._has_converged = metadata["converged"]
        traj._branching_history = metadata["branching_history"]
        traj._computed_steps = metadata["nstep_compute"]

        return traj

    @classmethod
    def restore_from_checkfile(
        cls,
        checkfile: Path,
        metadata_json: str,
        fmodel_t: type[ForwardModelBaseClass],
        parameters: dict[Any, Any],
        workdir: Path | None = None,
        frozen: bool = False,
    ) -> Trajectory:
        """Return a trajectory restored from an XML chkfile."""
        if not checkfile.exists():
            err_msg = f"Trajectory {checkfile} does not exist."
            _logger.exception(err_msg)
            raise FileNotFoundError

        rest_traj = Trajectory.init_from_metadata(metadata_json, fmodel_t, parameters, workdir, frozen)

        # Read in trajectory data
        tree = ET.parse(checkfile.absolute())
        root = tree.getroot()
        snapshots = root.find("snapshots")
        if snapshots is not None:
            for snap in snapshots:
                time, score, noise, state = read_xml_snapshot(snap)
                rest_traj._snaps.append(Snapshot(time, score, noise, state))

        # If the trajectory is frozen, that is all we need. Otherwise
        # handle sparse state, noise backlog and necessary fmodel initialization
        if rest_traj._fmodel:
            # Remove snapshots from the list until a state is available
            for k in range(len(rest_traj._snaps) - 1, -1, -1):
                if not rest_traj._snaps[k].has_state():
                    # Append the noise history to the backlog
                    rest_traj.noise_backlog.append(rest_traj._snaps[k].noise)
                    rest_traj._snaps.pop()
                else:
                    # Because the noise in the snapshot is the noise
                    # used to reach the next state, append the last to the backlog too
                    rest_traj.noise_backlog.append(rest_traj._snaps[k].noise)
                    break

            # Current step with python indexing, so remove 1
            rest_traj.set_current_time_and_step(rest_traj._snaps[-1].time, len(rest_traj._snaps) - 1)

            # Ensure everything is set to start the time stepping loop
            rest_traj.setup_noise()
            rest_traj._fmodel.set_current_state(rest_traj._snaps[-1].state)

            # Enable the model to perform tweaks
            # after a trajectory restore
            rest_traj._fmodel.post_trajectory_restore_hook(len(rest_traj._snaps) - 1, rest_traj.current_time())

        return rest_traj

    @classmethod
    def branch_from_trajectory(
        cls,
        from_traj: Trajectory,
        rst_traj: Trajectory,
        score: float,
        new_weight: float,
    ) -> Trajectory:
        """Create a new trajectory.

        Loading the beginning of a provided trajectory
        for all entries with score below a given score.
        This effectively branches the trajectory.

        Although the rst_traj is provided as an argument, it is
        only used to set metadata of the branched trajectory.

        Args:
            from_traj: an already existing trajectory to restart from
            rst_traj: the trajectory being restarted
            score: a threshold score
            new_weight: the weight of the child trajectory
        """
        # Check for empty trajectory
        if len(from_traj._snaps) == 0:
            tid, nb = get_index_from_id(rst_traj.idstr())
            new_workdir = Path(rst_traj.get_workdir().parents[0] / form_trajectory_id(tid, nb + 1))
            fmodel_t = type(from_traj._fmodel) if from_traj._fmodel else None
            rest_traj = Trajectory(
                traj_id=rst_traj.id(),
                weight=new_weight,
                fmodel_t=fmodel_t,
                parameters=from_traj._parameters_full,
                workdir=new_workdir,
            )
            rest_traj.set_checkfile(Path(rst_traj.get_checkfile().parents[0] / f"{rest_traj.idstr()}.xml"))
            return rest_traj

        # To ensure that TAMS converges, branching occurs on
        # the first snapshot with a score *strictly* above the target
        # Traverse the trajectory until a snapshot with a score >
        # the target is encountered
        high_score_idx = 0
        last_snap_with_state = 0
        while from_traj._snaps[high_score_idx].score <= score:
            high_score_idx += 1
            if from_traj._snaps[high_score_idx].has_state():
                last_snap_with_state = high_score_idx

        # Init empty trajectory
        tid, nb = get_index_from_id(rst_traj.idstr())
        new_workdir = Path(rst_traj.get_workdir().parents[0] / form_trajectory_id(tid, nb + 1))
        fmodel_t = type(from_traj._fmodel) if from_traj._fmodel else None
        rest_traj = Trajectory(
            traj_id=rst_traj.id(),
            weight=new_weight,
            fmodel_t=fmodel_t,
            parameters=from_traj._parameters_full,
            workdir=new_workdir,
        )
        rest_traj._branching_history = rst_traj._branching_history
        rest_traj._branching_history.append(from_traj.id())
        rest_traj.set_checkfile(Path(rst_traj.get_checkfile().parents[0] / f"{rest_traj.idstr()}.xml"))

        # If ancestor already have a backlog,
        # prepend it if the state id matches
        if last_snap_with_state == from_traj.get_last_state_id() and len(from_traj.noise_backlog) > 0:
            rest_traj.noise_backlog = rest_traj.noise_backlog + list(reversed(from_traj.noise_backlog))

        # Append snapshots, up to high_score_idx + 1 to
        # ensure > behavior
        for k in range(high_score_idx + 1):
            if k < last_snap_with_state:
                rest_traj._snaps.append(from_traj._snaps[k])
            elif k == last_snap_with_state:
                rest_traj._snaps.append(from_traj._snaps[k])
                rest_traj.noise_backlog.append(from_traj._snaps[k].noise)
            else:
                rest_traj.noise_backlog.append(from_traj._snaps[k].noise)

        # Reverse the backlog to ensure correct order
        rest_traj.noise_backlog.reverse()

        # Update trajectory metadata
        rest_traj._t_cur = rest_traj._snaps[-1].time
        rest_traj._step = len(rest_traj._snaps) - 1
        if rest_traj._fmodel:
            rest_traj.setup_noise()
            rest_traj._fmodel.set_current_state(rest_traj._snaps[-1].state)
            rest_traj.update_metadata()

            # Enable the model to perform tweaks
            # after a trajectory restart
            rest_traj._fmodel.post_trajectory_branching_hook(len(rest_traj._snaps) - 1, rest_traj._t_cur)

        return rest_traj

    def store(self, traj_file: Path | None = None) -> None:
        """Store the trajectory data to an XML chkfile."""
        root = ET.Element(self.idstr())
        root.append(dict_to_xml("params", self._parameters_full["trajectory"]))
        snaps_xml = ET.SubElement(root, "snapshots")
        for k in range(len(self._snaps)):
            snaps_xml.append(
                make_xml_snapshot(
                    k,
                    self._snaps[k].time,
                    self._snaps[k].score,
                    self._snaps[k].noise,
                    self._snaps[k].state,
                ),
            )
        tree = ET.ElementTree(root)
        ET.indent(tree, space="\t", level=0)

        if traj_file is not None:
            tree.write(traj_file.as_posix())
        else:
            tree.write(self._checkFile.as_posix())

    def set_weight(self, weight: float) -> None:
        """Set the trajectory weight.

        Args:
            weight: the (new) trajectory weight in the ensemble
        """
        self._weight = weight

    def update_metadata(self) -> None:
        """Update trajectory score/ending metadata.

        Update the maximum of the score function over the trajectory
        as well as the bool values for has_converged and has_ended.
        """
        new_score_max = 0.0
        for snap in self._snaps:
            new_score_max = max(new_score_max, snap.score)
        self._score_max = new_score_max
        if self._fmodel:
            self._has_converged = self._fmodel.check_convergence(
                self._step, self._t_cur, self._score_max, self._convergedVal
            )
        else:
            self._has_converged = False
        if self._t_cur >= self._t_end or self._has_converged:
            self._has_ended = True

    def set_current_time_and_step(self, time: float, step: int) -> None:
        """Set the current time and step."""
        self._t_cur = time
        self._step = step

    def current_time(self) -> float:
        """Return the current trajectory time."""
        return self._t_cur

    def step_size(self) -> float:
        """Return the time step size."""
        return self._dt

    def score_max(self) -> float:
        """Return the maximum of the score function."""
        return self._score_max

    def is_converged(self) -> bool:
        """Return True for converged trajectory."""
        return self._has_converged

    def has_ended(self) -> bool:
        """Return True for terminated trajectory."""
        return self._has_ended

    def has_started(self) -> bool:
        """Return True if computation has started."""
        return self._t_cur > 0.0

    def get_checkfile(self) -> Path:
        """Return the trajectory check file name."""
        return self._checkFile

    def get_time_array(self) -> npt.NDArray[np.number]:
        """Return the trajectory time instants."""
        times = np.zeros(len(self._snaps))
        for k in range(len(self._snaps)):
            times[k] = self._snaps[k].time
        return times

    def get_score_array(self) -> npt.NDArray[np.number]:
        """Return the trajectory scores."""
        scores = np.zeros(len(self._snaps))
        for k in range(len(self._snaps)):
            scores[k] = self._snaps[k].score
        return scores

    def get_noise_array(self) -> npt.NDArray[Any]:
        """Return the trajectory noises."""
        noises = np.zeros(len(self._snaps), dtype=type(self._snaps[0].noise))
        for k in range(len(self._snaps)):
            noises[k] = self._snaps[k].noise
        return noises

    def get_state_list(self) -> list[tuple[int, Any]]:
        """Return a list of states and associated indices.

        Returns:
            A list of tuples with index and states
        """
        return [(k, self._snaps[k].state) for k in range(len(self._snaps)) if self._snaps[k].has_state()]

    def get_length(self) -> int:
        """Return the trajectory length."""
        return len(self._snaps)

    def get_nbranching(self) -> int:
        """Return the number of branching events."""
        return len(self._branching_history)

    def get_computed_steps_count(self) -> int:
        """Return the number of compute steps taken."""
        return self._computed_steps

    def get_last_state(self) -> Any | None:
        """Return the last state in the trajectory."""
        for snap in reversed(self._snaps):
            if snap.has_state():
                return snap.state

        return None

    def get_last_state_id(self) -> int | None:
        """Return the id of the last state in the trajectory."""
        for s in reversed(range(len(self._snaps))):
            if self._snaps[s].has_state():
                return s

        return None

    def serialize_metadata_json(self) -> str:
        """Return a json string with metadata.

        Returns:
            A json string with the trajectory metadata
        """
        return json.dumps(
            {
                "id": self.id(),
                "weight": self._weight,
                "t_end": self._t_end,
                "t_cur": self._t_cur,
                "dt": self._dt,
                "ended": bool(self._has_ended),
                "converged": bool(self._has_converged),
                "score_max": self._score_max,
                "length": self.get_length(),
                "nstep_compute": self._computed_steps,
                "branching_history": self._branching_history,
            },
            default=str,
        )

    @classmethod
    def deserialize_metadata(cls, json_str: str) -> dict[str, Any]:
        """Load a json string into a properly typed metadata dict.

        Returns:
            A dictionary with the metadata
        """
        mstr = json.loads(json_str)
        return {
            "id": int(mstr["id"]),
            "weight": float(mstr["weight"]),
            "t_end": float(mstr["t_end"]),
            "t_cur": float(mstr["t_cur"]),
            "dt": float(mstr["dt"]),
            "ended": mstr["ended"],
            "converged": mstr["converged"],
            "score_max": float(mstr["score_max"]),
            "length": int(mstr["length"]),
            "nstep_compute": int(mstr["nstep_compute"]),
            "branching_history": cast("list[int]", mstr["branching_history"]),
        }

    def delete(self) -> None:
        """Clear the trajectory on-disk data."""
        self._checkFile.unlink()
        if self._workdir.exists():
            shutil.rmtree(self._workdir)



"""Tests for the pytams.utils class."""

import pytest
from pytams.utils import get_module_local_import
from tests.models import SimpleFModel


def test_get_local_import():
    """Test get local import function."""
    mods = get_module_local_import(SimpleFModel.__module__)
    assert len(mods) == 0


def test_get_local_import_fail():
    """Test get local import function with dummy module."""
    with pytest.raises(ValueError):
        _ = get_module_local_import("dummy_module_name")

def test_get_local_import_fail_stdlib():
    """Test get local import function with a module from stdlib."""
    with pytest.raises(FileNotFoundError):
        _ = get_module_local_import("time")



"""A set of utility for XML serializing."""

import ast
import logging
import xml.etree.ElementTree as ET
from collections.abc import Callable
from collections.abc import Generator
from contextlib import contextmanager
from datetime import datetime
from typing import Any
import numpy as np

_logger = logging.getLogger(__name__)


@contextmanager
def oneliner_ndarray() -> Generator[Any, None, None]:
    """Force ndarray print on a single line temporarily."""
    oldoptions = np.get_printoptions()
    np.set_printoptions(threshold=np.iinfo(np.int32).max, linewidth=np.iinfo(np.int32).max, precision=12)
    yield
    np.set_printoptions(**oldoptions)


class XMLUtilsError(Exception):
    """Exception class for the xmlutils."""


def manual_cast_snapshot(elem: ET.Element) -> Any:
    """Manually cast XML snapshot state."""
    if not elem.text:
        return elem.tag, None

    return elem.tag, manual_cast_str(elem.attrib["state_type"], elem.text)


def manual_cast_snapshot_noise(elem: ET.Element) -> Any:
    """Manually cast XML snapshot noise."""
    return elem.tag, manual_cast_str(elem.attrib["noise_type"], elem.attrib["noise"])


def manual_cast(elem: ET.Element) -> Any:
    """Manually cast XML elements reads."""
    if not elem.text:
        error_msg = f"Unable to parse XML element {elem.tag} since it is empty"
        _logger.exception(error_msg)
        raise XMLUtilsError(error_msg)

    return elem.tag, manual_cast_str(elem.attrib["type"], elem.text)


# Plain old data type cast in map
POD_cast_dict: dict[str, Callable[..., Any]] = {
    "int": int,
    "float": float,
    "float32": float,
    "float64": np.float64,
    "complex": complex,
    "str": str,
    "str_": str,
    "dict": ast.literal_eval,
    "tuple": ast.literal_eval,
    "bool": lambda elem_text: bool(elem_text == "True"),
    "bool_": lambda elem_text: bool(elem_text == "True"),
}


def manual_cast_str(type_str: str, elem_text: str) -> Any:
    """Manually cast from strings."""
    try:
        casted_elem = POD_cast_dict[type_str](elem_text)
    except KeyError as e:
        if type_str == "ndarray[float]":
            stripped_text = elem_text.replace("[", "").replace("]", "").replace("  ", " ")
            casted_elem = np.fromstring(stripped_text, dtype=float, sep=" ")
        elif type_str == "ndarray[int]":
            stripped_text = elem_text.replace("[", "").replace("]", "").replace("  ", " ")
            casted_elem = np.fromstring(stripped_text, dtype=int, sep=" ")
        elif type_str == "ndarray[bool]":
            stripped_text = elem_text.replace("[ ", " ").replace("]", "").replace("  ", " ")
            npofstr = np.array(list(stripped_text.lstrip().split(" ")), dtype=object)
            casted_elem = npofstr == "True"
        elif type_str == "datetime":
            casted_elem = datetime.strptime(elem_text, "%Y-%m-%d %H:%M:%S.%f%z")
        elif type_str == "None":
            casted_elem = None
        else:
            err_msg = f"Type {type_str} not handled by manual_cast !"
            _logger.exception(err_msg)
            raise XMLUtilsError(err_msg) from e
    return casted_elem


def dict_to_xml(tag: str, d: dict[Any, Any]) -> ET.Element:
    """Return an Element from a dictionary.

    Args:
        tag: a root tag
        d: a dictionary
    """
    elem = ET.Element(tag)
    for key, val in d.items():
        # Append an Element
        child = ET.Element(key)
        child.attrib["type"] = get_val_type(val)
        child.text = str(val)
        elem.append(child)

    return elem


def xml_to_dict(elem: ET.Element | None) -> dict[Any, Any]:
    """Return an dictionary an Element.

    Args:
        elem: an etree element

    Return:
        a dictionary containing the element entries
    """
    if elem is None:
        error_msg = "Unable to parse XML element to dict since 'None' was passed"
        _logger.exception(error_msg)
        raise XMLUtilsError(error_msg)

    d = {}
    for child in elem:
        tag, entry = manual_cast(child)
        d[tag] = entry

    return d


def get_val_type(val: Any) -> str:
    """Return the type of val.

    Args:
        val: a value

    Return:
        val type
    """
    base_type = type(val).__name__
    if base_type == "ndarray":
        if val.dtype == "float64":
            base_type = base_type + "[float]"
        elif val.dtype == "int64":
            base_type = base_type + "[int]"
        elif val.dtype == "bool":
            base_type = base_type + "[bool]"
        return base_type

    return base_type


def new_element(key: str, val: Any) -> ET.Element:
    """Return an Element from two args.

    Args:
        key: the element key
        val: the element value

    Return:
        an ElementTree element
    """
    elem = ET.Element(key)
    elem.attrib["type"] = get_val_type(val)
    with oneliner_ndarray():
        elem.text = str(val)

    return elem


def make_xml_snapshot(idx: int, time: float, score: float, noise: Any, state: Any) -> ET.Element:
    """Return a snapshot in XML element format.

    Args:
        idx: snapshot index
        time: the time stamp
        score: the snapshot score function
        noise: the stochastic noise
        state: the associated state
    """
    elem = ET.Element(f"Snap_{idx:07d}")
    elem.attrib["time"] = str(time)
    elem.attrib["score"] = str(score)
    elem.attrib["noise_type"] = get_val_type(noise)
    with oneliner_ndarray():
        elem.attrib["noise"] = str(noise)
    if state is None:
        elem.attrib["state_type"] = "None"
    else:
        elem.attrib["state_type"] = get_val_type(state)
        with oneliner_ndarray():
            elem.text = str(state)

    return elem


def read_xml_snapshot(snap: ET.Element) -> tuple[float, float, Any, Any]:
    """Return snapshot data from an XML snapshot element.

    Args:
        snap: an XML snapshot element
    """
    time = float(snap.attrib["time"])
    score = float(snap.attrib["score"])
    _, noise = manual_cast_snapshot_noise(snap)
    _, state = manual_cast_snapshot(snap)

    return time, score, noise, state



"""A class for the TAMS data as an SQL database using SQLAlchemy."""

from __future__ import annotations
import gc
import json
import logging
from pathlib import Path
from typing import cast
import numpy as np
import numpy.typing as npt
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import sessionmaker

_logger = logging.getLogger(__name__)


class Base(DeclarativeBase):
    """A base class for the tables."""


class Trajectory(Base):
    """A table storing the active trajectories."""

    __tablename__ = "trajectories"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    traj_file: Mapped[str] = mapped_column(nullable=False)
    t_metadata: Mapped[str] = mapped_column(default="", nullable=False)
    status: Mapped[str] = mapped_column(default="idle", nullable=False)


class ArchivedTrajectory(Base):
    """A table storing the archived trajectories."""

    __tablename__ = "archived_trajectories"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    traj_file: Mapped[str] = mapped_column(nullable=False)
    t_metadata: Mapped[str] = mapped_column(default="", nullable=False)


class SplittingIterations(Base):
    """A table storing the splitting iterations."""

    __tablename__ = "splitting_iterations"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    split_id: Mapped[int] = mapped_column(nullable=False)
    bias: Mapped[int] = mapped_column(nullable=False)
    weight: Mapped[str] = mapped_column(nullable=False)
    discarded_traj_ids: Mapped[str] = mapped_column(nullable=False)
    ancestor_traj_ids: Mapped[str] = mapped_column(nullable=False)
    min_vals: Mapped[str] = mapped_column(nullable=False)
    min_max: Mapped[str] = mapped_column(nullable=False)
    status: Mapped[str] = mapped_column(default="locked", nullable=False)


valid_statuses = ["locked", "idle", "completed"]


class SQLFile:
    """An SQL file.

    Allows atomic access to an SQL database from all
    the workers.

    Note: TAMS works with Python indexing starting at 0,
    while SQL indexing starts at 1. Trajectory ID is
    updated accordingly when accessing/updating the DB.

    Attributes:
        _file_name : The file name
    """

    def __init__(self, file_name: str, in_memory: bool = False, ro_mode: bool = False) -> None:
        """Initialize the file.

        Args:
            file_name : The file name
            in_memory: a bool to trigger in-memory creation
            ro_mode: a bool to trigger read-only access to the database
        """
        self._file_name = "" if in_memory else file_name

        # URI mode requires absolute path
        file_path = Path(file_name).absolute().as_posix()
        if in_memory:
            self._engine = create_engine("sqlite:///:memory:", echo=False)
        else:
            self._engine = (
                create_engine(f"sqlite:///file:{file_path}?mode=ro&uri=true", echo=False)
                if ro_mode
                else create_engine(f"sqlite:///{file_path}", echo=False)
            )
        self._Session = sessionmaker(bind=self._engine)
        self._init_db()

    def _init_db(self) -> None:
        """Initialize the tables of the file.

        Raises:
            RuntimeError : If a connection to the DB could not be acquired
        """
        try:
            Base.metadata.create_all(self._engine)
        except SQLAlchemyError:
            err_msg = "Failed to initialize DB schema"
            _logger.exception(err_msg)
            raise

    def name(self) -> str:
        """Access the DB file name.

        Returns:
            the database name, empty string if in-memory
        """
        return self._file_name

    def add_trajectory(self, traj_file: str, metadata: str) -> None:
        """Add a new trajectory to the DB.

        Args:
            traj_file : The trajectory file of that trajectory
            metadata: a json representation of the traj metadata

        Raises:
            SQLAlchemyError if the DB could not be accessed
        """
        session = self._Session()
        try:
            new_traj = Trajectory(traj_file=traj_file, t_metadata=metadata)
            session.add(new_traj)
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to add trajectory")
            raise
        finally:
            session.close()

    def update_trajectory(self, traj_id: int, traj_file: str, metadata: str) -> None:
        """Update a given trajectory data in the DB.

        Args:
            traj_id : The trajectory id
            traj_file : The new trajectory file of that trajectory
            metadata: a json representation of the traj metadata

        Raises:
            SQLAlchemyError if the DB could not be accessed
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).one()
            traj.traj_file = mapped_column(traj_file)
            traj.t_metadata = metadata
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            err_msg = f"Failed to update trajectory {traj_id}"
            _logger.exception(err_msg)
            raise
        finally:
            session.close()

    def update_trajectory_weight(self, traj_id: int, weight: float) -> None:
        """Update a given trajectory weight in the DB.

        Args:
            traj_id : The trajectory id
            weight: the new trajectory weight

        Raises:
            SQLAlchemyError if the DB could not be accessed
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).one()
            metadata_d = json.loads(traj.t_metadata)
            metadata_d["weight"] = str(weight)
            traj.t_metadata = mapped_column(json.dumps(metadata_d))
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            err_msg = f"Failed to update trajectory {traj_id} weight"
            _logger.exception(err_msg)
            raise
        finally:
            session.close()

    def lock_trajectory(self, traj_id: int, allow_completed_lock: bool = False) -> bool:
        """Set the status of a trajectory to "locked" if possible.

        Args:
            traj_id : The trajectory id
            allow_completed_lock : Allow to lock a "completed" trajectory

        Return:
            True if the trajectory was successfully locked, False otherwise

        Raises:
            ValueError if the trajectory with the given id does not exist
            SQLAlchemyError if the DB could not be accessed
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).with_for_update().one_or_none()

            if traj:
                allowed_status = ["idle", "completed"] if allow_completed_lock else ["idle"]
                if traj.status in allowed_status:
                    traj.status = "locked"
                    session.commit()
                    return True
                return False

            err_msg = f"Trajectory {traj_id} does not exist"
            _logger.error(err_msg)
            raise ValueError(err_msg)

        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to lock trajectory")
            raise
        finally:
            session.close()

    def mark_trajectory_as_completed(self, traj_id: int) -> None:
        """Set the status of a trajectory to "completed" if possible.

        Args:
            traj_id : The trajectory id

        Raises:
            ValueError if the trajectory with the given id does not exist
            SQLAlchemyError if the DB could not be accessed
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).one_or_none()
            if traj:
                if traj.status in ["locked"]:
                    traj.status = "completed"
                    session.commit()
                else:
                    warn_msg = f"Attempting to mark completed Trajectory {traj_id} already in status {traj.status}."
                    _logger.warning(warn_msg)
            else:
                err_msg = f"Trajectory {traj_id} does not exist"
                _logger.error(err_msg)
                raise ValueError(err_msg)
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to mark trajectory as completed")
            raise
        finally:
            session.close()

    def release_trajectory(self, traj_id: int) -> None:
        """Set the status of a trajectory to "idle" if possible.

        Args:
            traj_id : The trajectory id

        Raises:
            ValueError if the trajectory with the given id does not exist
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).one_or_none()
            if traj:
                if traj.status in ["locked"]:
                    traj.status = "idle"
                    session.commit()
                else:
                    warn_msg = f"Attempting to release Trajectory {traj_id} already in status {traj.status}."
                    _logger.warning(warn_msg)
            else:
                err_msg = f"Trajectory {traj_id} does not exist"
                _logger.error(err_msg)
                raise ValueError(err_msg)
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to release trajectory")
            raise
        finally:
            session.close()

    def get_trajectory_count(self) -> int:
        """Get the number of trajectories in the DB.

        Returns:
            The number of trajectories
        """
        session = self._Session()
        try:
            return session.query(Trajectory).count()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to count the number of trajectories")
            raise
        finally:
            session.close()

    def fetch_trajectory(self, traj_id: int) -> tuple[str, str]:
        """Get the trajectory file of a trajectory.

        Args:
            traj_id : The trajectory id

        Return:
            The trajectory file

        Raises:
            ValueError if the trajectory with the given id does not exist
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(Trajectory).filter(Trajectory.id == db_id).one_or_none()
            if traj:
                tfile: str = traj.traj_file
                metadata_str: str = traj.t_metadata
                return tfile, metadata_str

            err_msg = f"Trajectory {traj_id} does not exist"
            _logger.error(err_msg)
            raise ValueError(err_msg)
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to fetch trajectory")
            raise
        finally:
            session.close()

    def release_all_trajectories(self) -> None:
        """Release all trajectories in the DB."""
        session = self._Session()
        try:
            session.query(Trajectory).update({"status": "idle"})
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to release all trajectories")
        finally:
            session.close()

    def archive_trajectory(self, traj_file: str, metadata: str) -> None:
        """Add a new trajectory to the archive container.

        Args:
            traj_file : The trajectory file of that trajectory
            metadata: a json representation of the traj metadata
        """
        session = self._Session()
        try:
            new_traj = ArchivedTrajectory(traj_file=traj_file, t_metadata=metadata)
            session.add(new_traj)
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to archive trajectory")
        finally:
            session.close()

    def fetch_archived_trajectory(self, traj_id: int) -> tuple[str, str]:
        """Get the trajectory file of a trajectory in the archive.

        Args:
            traj_id : The trajectory id

        Return:
            The trajectory file

        Raises:
            ValueError if the trajectory with the given id does not exist
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = traj_id + 1
            traj = session.query(ArchivedTrajectory).filter(ArchivedTrajectory.id == db_id).one_or_none()
            if traj:
                tfile: str = traj.traj_file
                metadata_str: str = traj.t_metadata
                return tfile, metadata_str

            err_msg = f"Trajectory {traj_id} does not exist"
            _logger.error(err_msg)
            raise ValueError(err_msg)
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to fetch archived trajectory")
            raise
        finally:
            session.close()

    def get_archived_trajectory_count(self) -> int:
        """Get the number of trajectories in the archive.

        Returns:
            The number of trajectories
        """
        session = self._Session()
        try:
            return session.query(ArchivedTrajectory).count()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to count the number of archived trajectories")
            raise
        finally:
            session.close()

    def clear_archived_trajectories(self) -> int:
        """Delete the content of the archived traj table.

        Returns:
            The number of entries deleted
        """
        session = self._Session()
        try:
            ndelete = session.query(ArchivedTrajectory).delete()
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to clear archived trajectories")
            raise
        else:
            return ndelete
        finally:
            session.close()

    def add_splitting_data(
        self,
        k: int,
        bias: int,
        weight: float,
        discarded_ids: list[int],
        ancestor_ids: list[int],
        min_vals: list[float],
        min_max: list[float],
    ) -> None:
        """Add a new splitting data to the DB.

        Args:
            k : The splitting iteration index
            bias : The number of restarted trajectories
            weight : Weight of the ensemble at the current iteration
            discarded_ids : The list of discarded trajectory ids
            ancestor_ids : The list of trajectories used to restart
            min_vals : The list of minimum values
            min_max : The score minimum and maximum values
        """
        session = self._Session()
        try:
            new_split = SplittingIterations(
                split_id=k,
                bias=bias,
                weight=str(weight),
                discarded_traj_ids=json.dumps(discarded_ids),
                ancestor_traj_ids=json.dumps(ancestor_ids),
                min_vals=json.dumps(min_vals),
                min_max=json.dumps(min_max),
            )
            session.add(new_split)
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to add splitting data")
            raise
        finally:
            session.close()

    def update_splitting_data(
        self,
        k: int,
        bias: int,
        weight: float,
        discarded_ids: list[int],
        ancestor_ids: list[int],
        min_vals: list[float],
        min_max: list[float],
    ) -> None:
        """Update the last splitting data row to the DB.

        Args:
            k : The splitting iteration index
            bias : The number of restarted trajectories
            weight : Weight of the ensemble at the current iteration
            discarded_ids : The list of discarded trajectory ids
            ancestor_ids : The list of trajectories used to restart
            min_vals : The list of minimum values
            min_max : The score minimum and maximum values
        """
        session = self._Session()
        try:
            dset = session.query(SplittingIterations).order_by(SplittingIterations.id.desc()).first()
            if dset:
                dset.split_id = k
                dset.bias = bias
                dset.weight = str(weight)
                dset.discarded_traj_ids = json.dumps(discarded_ids)
                dset.ancestor_traj_ids = json.dumps(ancestor_ids)
                dset.min_vals = json.dumps(min_vals)
                dset.min_max = json.dumps(min_max)
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to update the last splitting data")
            raise
        finally:
            session.close()

    def mark_last_iteration_as_completed(self) -> None:
        """Mark the last splitting iteration as complete.

        By default, iteration data append to the SQL table with a state "locked"
        to indicate an iteration being worked on. Upon completion, mark it as
        "completed" otherwise the iteration is considered incomplete, i.e.
        interrupted by some error or wall clock limit.
        """
        session = self._Session()
        try:
            iteration = session.query(SplittingIterations).order_by(SplittingIterations.id.desc()).first()
            if iteration:
                iteration.status = "completed"
                session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to mark splitting iteration as completed")
            raise
        finally:
            session.close()

    def get_k_split(self) -> int:
        """Get the current splitting iteration counter.

        Returns:
            The ksplit from the last entry in the SplittingIterations table
        """
        session = self._Session()
        try:
            last_split = session.query(SplittingIterations).order_by(SplittingIterations.id.desc()).first()
            if last_split:
                return last_split.split_id + last_split.bias
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to query k_split !")
            raise
        else:
            return 0
        finally:
            session.close()

    def get_iteration_count(self) -> int:
        """Get the number of splitting iteration stored.

        Returns:
            The length of the SplittingIterations table
        """
        session = self._Session()
        try:
            return session.query(SplittingIterations).count()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to count the number of splitting iteration stored")
            raise
        finally:
            session.close()

    def fetch_splitting_data(
        self, k_id: int
    ) -> tuple[int, int, float, list[int], list[int], list[float], list[float], str] | None:
        """Get the splitting iteration data for a given iteration.

        Args:
            k_id : The iteration id

        Return:
            The splitting iteration data

        Raises:
            ValueError if the splitting iteration with the given id does not exist
        """
        session = self._Session()
        try:
            # SQL indexing starts at 1, adjust ID
            db_id = k_id + 1
            split = session.query(SplittingIterations).filter(SplittingIterations.id == db_id).one_or_none()
            if split:
                return (
                    int(split.split_id),
                    int(split.bias),
                    float(split.weight),
                    cast("list[int]", json.loads(split.discarded_traj_ids)),
                    cast("list[int]", json.loads(split.ancestor_traj_ids)),
                    cast("list[float]", json.loads(split.min_vals)),
                    cast("list[float]", json.loads(split.min_max)),
                    split.status,
                )

            err_msg = f"Splitting iteration {k_id} does not exist"
            _logger.error(err_msg)
            raise ValueError(err_msg)
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to fetch splitting iteration data")
            raise
        finally:
            session.close()

    def get_ongoing(self) -> list[int] | None:
        """Get the list of ongoing trajectories if any.

        Returns:
            Either a list trajectories or None if nothing was left to do
        """
        session = self._Session()
        try:
            last_split = session.query(SplittingIterations).order_by(SplittingIterations.id.desc()).first()
            if last_split and last_split.status == "locked":
                return cast("list[int]", json.loads(last_split.discarded_traj_ids))
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to query the list of ongoing trajectories !")
            raise
        else:
            return None
        finally:
            session.close()

    def get_weights(self) -> npt.NDArray[np.number]:
        """Read the weights from the database.

        Returns:
            the weight for each splitting iteration as a numpy array
        """
        session = self._Session()
        try:
            return np.array(
                [r[0] for r in session.query(SplittingIterations).with_entities(SplittingIterations.weight).all()],
                dtype="float32",
            )
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to access the weights !")
            raise
        finally:
            session.close()

    def get_biases(self) -> npt.NDArray[np.number]:
        """Read the biases from the database.

        Returns:
            the bias for each splitting iteration as a numpy array
        """
        session = self._Session()
        try:
            return np.array(
                [r[0] for r in session.query(SplittingIterations).with_entities(SplittingIterations.bias).all()],
                dtype="int",
            )
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to access the biases !")
            raise
        finally:
            session.close()

    def get_minmax(self) -> npt.NDArray[np.number]:
        """Read the min/max from the database.

        Returns:
            the 2D Numpy array with k_index, min, max
        """
        session = self._Session()
        try:
            return np.array(
                [
                    [r[0], json.loads(r[1])[0], json.loads(r[1])[1]]
                    for r in session.query(SplittingIterations)
                    .with_entities(SplittingIterations.split_id, SplittingIterations.min_max)
                    .all()
                ],
                dtype="float32",
            )
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to access the biases !")
            raise
        finally:
            session.close()

    def clear_splitting_data(self) -> int:
        """Delete the content of the splitting data table.

        Returns:
            The number of entries deleted
        """
        session = self._Session()
        try:
            ndelete = session.query(SplittingIterations).delete()
            session.commit()
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to clear splitting iterations data.")
            raise
        else:
            return ndelete
        finally:
            session.close()

    def dump_file_json(self, json_file: str | None = None) -> None:
        """Dump the content of the trajectory table to a json file.

        Args:
            json_file: an optional file name (or path) to dump the data to
        """
        db_data = {}
        session = self._Session()
        try:
            db_data["trajectories"] = {
                traj.id - 1: {"file": traj.traj_file, "status": traj.status, "metadata": traj.t_metadata}
                for traj in session.query(Trajectory).all()
            }
            db_data["archived_trajectories"] = {
                traj.id - 1: {"file": traj.traj_file, "metadata": traj.t_metadata}
                for traj in session.query(ArchivedTrajectory).all()
            }
            db_data["splitting_data"] = {
                split.id: {
                    "k": str(split.split_id),
                    "bias": str(split.bias),
                    "weight": split.weight,
                    "min_max_start": json.loads(split.min_max),
                    "discarded_ids": json.loads(split.discarded_traj_ids),
                    "ancestor_ids": json.loads(split.ancestor_traj_ids),
                    "min_vals": json.loads(split.min_vals),
                    "status": split.status,
                }
                for split in session.query(SplittingIterations).all()
            }
        except SQLAlchemyError:
            session.rollback()
            _logger.exception("Failed to query the content of the DB")
            raise
        finally:
            session.close()

        json_path = Path(json_file) if json_file else Path(f"{Path(self._file_name).stem}.json")
        with json_path.open("w") as f:
            json.dump(db_data, f, indent=2)

    def __del__(self) -> None:
        """Explicit delete function.

        On windows, the SQL file is locked.
        """
        del self._Session
        self._engine.dispose()
        gc.collect()



"""Tests for the pytams.database class."""

import shutil
from pathlib import Path
import pytest
import toml
from pytams.database import Database
from pytams.tams import TAMS
from tests.models import DoubleWellModel


def test_failed_db_init_no_ntraj():
    """Test init of TDB from scratch missing argument."""
    fmodel = DoubleWellModel
    params_load_db = {}
    with pytest.raises(ValueError):
        Database(fmodel, params_load_db)

def test_failed_db_init_no_nsplit():
    """Test init of TDB failing missing nsplit."""
    fmodel = DoubleWellModel
    params_load_db = {}
    with pytest.raises(ValueError):
        Database(fmodel, params_load_db, ntraj=10)

def test_wrong_format():
    """Test init of TDB with unsupported format."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb", "format": "WRONG"}}
    with pytest.raises(ValueError):
        _ = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)

def test_load_missing_tdb():
    """Test failed load database."""
    with pytest.raises(FileNotFoundError):
        _ = Database.load(Path("dwTestNonExistent.tdb"))

def test_init_empty_tdb_inmemory():
    """Test init database."""
    fmodel = DoubleWellModel
    params_load_db = {}
    tdb = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)
    assert tdb.name() == "TAMS_DoubleWellModel"

def test_init_empty_tdb():
    """Test init database on disk."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)
    assert tdb.name() == "dwTest.tdb"
    # Necessary on Windows
    del tdb
    shutil.rmtree("dwTest.tdb")

def test_reinit_empty_tdb():
    """Test init database on disk."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTestDouble.tdb"}}
    tdb = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)
    # Necessary on Windows
    del tdb
    params_load_db = {"database": {"path": "dwTestDouble.tdb", "restart": True}}
    tdb = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)
    del tdb
    ndb = 0
    for folder in Path("./").iterdir():
        if "dwTestDouble" in str(folder):
            shutil.rmtree(folder)
            ndb += 1
    assert ndb == 2

def test_init_and_load_empty_tdb():
    """Test init database on disk."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db, ntraj=10, nsplititer=100)
    tdb_path = Path(tdb.path())
    assert tdb.name() == "dwTest.tdb"
    del tdb
    tdb = Database.load(tdb_path)
    del tdb
    shutil.rmtree("dwTest.tdb")

@pytest.mark.dependency(name="genDB")
def test_generate_and_load_tdb():
    """Test generation of TDB and loading the TDB."""
    fmodel = DoubleWellModel
    with Path("input.toml").open("w") as f:
        toml.dump(
            {
                "tams": {"ntrajectories": 50, "nsplititer": 200, "walltime": 500.0, "loglevel": "DEBUG"},
                "database": {"path": "dwTest.tdb"},
                "runner": {"type": "asyncio", "nworker_init": 2, "nworker_iter": 1},
                "model": {"noise_amplitude": 0.8},
                "trajectory": {"end_time": 10.0, "step_size": 0.01, "targetscore": 0.51},
            },
            f,
        )
    tams = TAMS(fmodel_t=fmodel, a_args=[])
    tams.compute_probability()

    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.info()
    assert tdb
    Path("input.toml").unlink(missing_ok=True)

@pytest.mark.dependency(depends=["genDB"])
def test_access_ensemble_length():
    """Test accessing database trajectory ensemble size."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    assert tdb.is_empty() is False

@pytest.mark.dependency(depends=["genDB"])
def test_access_ended_count():
    """Test accessing database trajectory metadata."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    assert tdb.count_ended_traj() == 50

@pytest.mark.dependency(depends=["genDB"])
def test_access_converged_count():
    """Test accessing database trajectory metadata."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    assert tdb.count_converged_traj() == 50

@pytest.mark.dependency(depends=["genDB"])
def test_access_pool_content():
    """Test accessing database SQL pool content."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.get_pool_db().dump_file_json("test.json")
    assert Path("./test.json").exists()
    Path("./test.json").unlink()

@pytest.mark.dependency(depends=["genDB"])
def test_copy_and_access():
    """Test copying the database and accessing it."""
    shutil.copytree("dwTest.tdb", "dwTestCopy.tdb")
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTestCopy.tdb"}}
    tdb = Database(fmodel, params_load_db)
    assert tdb.count_converged_traj() == 50
    del tdb
    shutil.rmtree("dwTestCopy.tdb")

@pytest.mark.dependency(depends=["genDB"])
def test_replace_traj_in_tdb():
    """Test replacing a trajectory in the database."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()

    traj_zero = tdb.get_traj(0)
    tdb.overwrite_traj(1, traj_zero)
    assert tdb.get_traj(1).idstr()[:10] == "traj000000"

@pytest.mark.dependency(depends=["genDB"])
def test_unknown_traj_access_in_tdb():
    """Test accessing a trajectory out-of-range."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()

    with pytest.raises(ValueError):
        _ = tdb.get_traj(10000)

@pytest.mark.dependency(depends=["genDB"])
def test_unknown_traj_overwrite_in_tdb():
    """Test overwriting a trajectory out-of-range."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()

    traj_zero = tdb.get_traj(0)
    with pytest.raises(ValueError):
        _ = tdb.overwrite_traj(10000, traj_zero)

@pytest.mark.dependency(depends=["genDB"])
def test_access_trajdata_in_tdb():
    """Test accessing a trajectory in the database."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()

    traj = tdb.get_traj(0)
    times = traj.get_time_array()
    scores = traj.get_score_array()
    noises = traj.get_noise_array()
    assert times.size > 0
    assert scores.size > 0
    assert noises.size > 0

@pytest.mark.dependency(depends=["genDB"])
def test_explore_tdb():
    """Test loading the TDB."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()
    tdb.plot_score_functions("test.png")
    Path("./test.png").unlink(missing_ok=False)

@pytest.mark.dependency(depends=["genDB"])
def test_explore_minmax_tdb():
    """Test loading the TDB."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db)
    tdb.load_data()
    tdb.plot_min_max_span(fname="test_minmax.png")
    Path("./test_minmax.png").unlink(missing_ok=False)

@pytest.mark.dependency(depends=["genDB"])
def test_explore_active_at_k():
    """Test getting the initial active set."""
    tdb = Database.load(Path("dwTest.tdb"))
    tdb.load_data(load_archived_trajectories=True)
    act_trajs = tdb.get_trajectory_active_at_k(0)
    assert len(act_trajs) == 50
    assert act_trajs[42].idstr() == "traj000042_0000"

@pytest.mark.dependency(depends=["genDB"])
def test_restore_tdb():
    """Test loading and restoring the TDB."""
    fmodel = DoubleWellModel
    params_load_db = {"database": {"path": "dwTest.tdb"}}
    tdb = Database(fmodel, params_load_db, read_only=False)
    tdb.load_data()
    tdb.reset_initial_ensemble_stage()
    assert tdb.k_split() == 0
    del tdb
    shutil.rmtree("dwTest.tdb")



"""Tests for the pytams.trajectory class."""

from math import isclose
from pathlib import Path
import pytest
from pytams.fmodel import ForwardModelBaseClass
from pytams.trajectory import Snapshot
from pytams.trajectory import Trajectory
from pytams.utils import moving_avg
from tests.models import DoubleWellModel
from tests.models import SimpleFModel


def test_init_snapshot():
    """Test initialization of a snapshot."""
    snap = Snapshot(0.1, 0.1, "Noisy", "State")
    assert snap.time == 0.1
    assert snap.has_state()


def test_init_snapshot_nostate():
    """Test initialization of a stateless snapshot."""
    snap = Snapshot(0.1, 0.1, "Noisy")
    assert not snap.has_state()


def test_init_missing_basic_inputs():
    """Test lack of minimal traj data in trajectory creation."""
    fmodel = ForwardModelBaseClass
    parameters = {}
    with pytest.raises(ValueError):
        _ = Trajectory(1, 1.0, fmodel, parameters)


def test_init_baseclasserror():
    """Test using base class fmodel during trajectory creation."""
    fmodel = ForwardModelBaseClass
    parameters = {"trajectory": {"end_time": 2.0, "step_size": 0.01}}
    with pytest.raises(TypeError):
        _ = Trajectory(1, 1.0, fmodel, parameters)


def test_init_blank_traj():
    """Test blank trajectory creation."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 2.0, "step_size": 0.01}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    assert t_test.id() == 1
    assert t_test.idstr() == "traj000001_0000"
    assert t_test.current_time() == 0.0
    assert t_test.score_max() == -1000000000000.0


def test_init_parametrized_traj():
    """Test parametrized trajectory creation."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 2.0, "step_size": 0.01, "targetscore": 0.25}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.set_workdir(Path())
    assert t_test.step_size() == 0.01


def test_restart_empty_traj():
    """Test (empty) trajectory restart."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 2.0, "step_size": 0.01}}
    from_traj = Trajectory(1, 0.5, fmodel, parameters)
    rst_traj = Trajectory(2, 0.5, fmodel, parameters)
    rst_test = Trajectory.branch_from_trajectory(from_traj, rst_traj, 0.1, 0.25)
    assert rst_test.current_time() == 0.0


def test_simple_model_traj():
    """Test trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.001, "targetscore": 0.25}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.01)
    assert isclose(t_test.score_max(), 0.1, abs_tol=1e-9)
    assert t_test.is_converged() is False
    t_test.advance()
    assert t_test.is_converged() is True


def test_branch_simple_model_traj():
    """Test branching a trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.0002, "targetscore": 0.45}}
    t_ancestor = Trajectory(1 ,0.5, fmodel, parameters)
    t_ancestor.advance()
    assert t_ancestor.get_computed_steps_count() == 201
    t_branched = Trajectory(2, 0.5, fmodel, parameters)
    t_branched = Trajectory.branch_from_trajectory(t_ancestor, t_branched, 0.1, 0.25)
    assert t_branched.get_computed_steps_count() == 0
    t_branched.advance()
    assert t_branched.get_computed_steps_count() == 150


def test_store_and_restore_simple_traj():
    """Test store and restoring trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.05, "step_size": 0.001, "targetscore": 0.25}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.02)
    assert isclose(t_test.score_max(), 0.2, abs_tol=1e-9)
    assert t_test.is_converged() is False
    chkfile = Path("./test.xml")
    t_test.store(chkfile)
    assert chkfile.exists() is True
    metadata = t_test.serialize_metadata_json()
    rst_test = Trajectory.restore_from_checkfile(chkfile, metadata, fmodel, parameters)
    assert isclose(rst_test.score_max(), 0.2, abs_tol=1e-9)
    rst_test.advance()
    assert rst_test.is_converged() is True
    chkfile.unlink(missing_ok=True)


def test_store_and_restore_frozen_simple_traj():
    """Test store and restoring frozen trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.05, "step_size": 0.001, "targetscore": 0.25}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.02)
    assert isclose(t_test.score_max(), 0.2, abs_tol=1e-9)
    assert t_test.is_converged() is False
    chkfile = Path("./test.xml")
    t_test.store(chkfile)
    assert chkfile.exists() is True
    metadata = t_test.serialize_metadata_json()
    rst_test = Trajectory.restore_from_checkfile(chkfile, metadata, fmodel, parameters, frozen=True)
    assert isclose(rst_test.score_max(), 0.2, abs_tol=1e-9)
    with pytest.raises(RuntimeError):
        rst_test.advance()
    with pytest.raises(RuntimeError):
        rst_test._one_step()
    chkfile.unlink(missing_ok=True)


def test_restart_simple_traj():
    """Test trajectory restart."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.001, "targetscore": 0.25}}
    from_traj = Trajectory(1, 0.5, fmodel, parameters)
    from_traj.advance(0.01)
    rst_traj = Trajectory(2, 0.5, fmodel, parameters)
    rst_test = Trajectory.branch_from_trajectory(from_traj, rst_traj, 0.05, 0.25)
    assert rst_test.current_time() == 0.006


def test_access_data_simple_traj():
    """Test trajectory data access."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.001, "targetscore": 0.25}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.01)
    assert t_test.get_length() == 11
    assert isclose(t_test.get_time_array()[-1], 0.01, abs_tol=1e-9)
    assert isclose(t_test.get_score_array()[-1], 0.1, abs_tol=1e-9)


def test_sparse_simple_traj():
    """Test a sparse trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.001, "targetscore": 0.25, "sparse_freq": 5}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.012)
    assert isclose(t_test.score_max(), 0.12, abs_tol=1e-9)
    assert t_test.is_converged() is False
    assert isclose(t_test.get_last_state(), 0.009, abs_tol=1e-9)
    t_test.advance()
    assert t_test.is_converged() is True
    assert isclose(t_test.get_last_state(), 0.024, abs_tol=1e-9)


def test_sparse_simple_traj_access_states():
    """Test a sparse trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.0002, "targetscore": 0.25, "sparse_freq": 5}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance()
    assert len(t_test.get_state_list()) == 26


def test_store_and_restart_sparse_simple_traj():
    """Test a sparse trajectory with simple model."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.04, "step_size": 0.001, "targetscore": 0.25, "sparse_freq": 5}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(0.013)
    assert isclose(t_test.score_max(), 0.13, abs_tol=1e-9)
    assert t_test.is_converged() is False
    chkfile = Path("./test.xml")
    t_test.store(chkfile)
    assert chkfile.exists() is True
    metadata = t_test.serialize_metadata_json()
    rst_test = Trajectory.restore_from_checkfile(chkfile, metadata, fmodel, parameters)
    rst_test.advance()
    assert rst_test.is_converged() is True
    chkfile.unlink()


def test_score_moving_average():
    """Test using a moving average on a score array."""
    fmodel = SimpleFModel
    parameters = {"trajectory": {"end_time": 0.9, "step_size": 0.0001, "targetscore": 0.95}}
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance()
    score = t_test.get_score_array()
    avg_score = moving_avg(score, 10)
    assert isclose(avg_score[0], 0.0045, abs_tol=1e-9)


def test_sparse_dw_traj_with_restore():
    """Test restore a sparse trajectory with DW model."""
    fmodel = DoubleWellModel
    parameters = {
        "tams": {"deterministic": True},
        "trajectory": {"end_time": 15.0, "step_size": 0.01, "targetscore": 0.95, "sparse_freq": 10},
        "model": {"noise_amplitude": 0.8},
    }
    t_test = Trajectory(1, 1.0, fmodel, parameters)
    t_test.advance(4.07)
    chkfile = Path("./test.xml")
    t_test.store(chkfile)
    assert isclose(t_test.score_max(), 0.5383998247480907, abs_tol=1e-9)
    assert not t_test.is_converged()
    metadata = t_test.serialize_metadata_json()
    rst_test = Trajectory.restore_from_checkfile(chkfile, metadata, fmodel, parameters, frozen=False)
    rst_test.advance()
    assert rst_test.score_max() > 0.95
    assert rst_test.is_converged()
    chkfile.unlink(missing_ok=True)


def test_sparse_dw_traj_with_branching():
    """Test branching a sparse trajectory with simple model."""
    fmodel = DoubleWellModel
    parameters = {
        "tams": {"deterministic": True},
        "trajectory": {"end_time": 2.0, "step_size": 0.01, "targetscore": 0.95, "sparse_freq": 10},
        "model": {"noise_amplitude": 0.3},
    }
    t_test = [Trajectory(1, 0.5, fmodel, parameters), Trajectory(2, 0.5, fmodel, parameters)]
    t_test[0].advance()
    t_test[1].advance()
    if t_test[0].score_max() > t_test[1].score_max():
        rst_idx = 1
        from_idx = 0
        rst_val = t_test[1].score_max()
    else:
        rst_idx = 0
        from_idx = 1
        rst_val = t_test[0].score_max()
    branched_test = Trajectory.branch_from_trajectory(t_test[from_idx], t_test[rst_idx], rst_val, 0.25)
    branched_test.advance()
    assert branched_test.score_max() > t_test[rst_idx].score_max()
